import threading, atexit, sys
from collections import namedtuple
import os.path

if sys.version_info[0] >= 3:
    from _thread import start_new_thread
else:
    from thread import start_new_thread

FrameInfo = namedtuple('FrameInfo', 'filename, name, f_trace')


def _atexit():
    sys.stderr.flush()
    sys.stdout.flush()


# Register the TEST SUCEEDED msg to the exit of the process.
atexit.register(_atexit)


def _iter_frame_info(frame):
    while frame is not None:
        yield FrameInfo(
            os.path.basename(frame.f_code.co_filename),
            frame.f_code.co_name,
            frame.f_trace.__name__ if frame.f_trace is not None else "None"
        )
        frame = frame.f_back


def check_frame_info(expected):
    found = list(_iter_frame_info(sys._getframe().f_back))

    def fail():
        raise AssertionError('Expected:\n%s\n\nFound:\n%s\n' % (
            '\n'.join(str(x) for x in expected),
            '\n'.join(str(x) for x in found)))

    for found_info, expected_info in  zip(found, expected):
        if found_info.filename != expected_info.filename or found_info.name != expected_info.name:
            fail()
            
        for f_trace in expected_info.f_trace.split('|'):
            if f_trace == found_info.f_trace:
                break
        else:
            fail()


def thread_func():
    if sys.version_info[0] >= 3:
        check_frame_info([
            FrameInfo(filename='_debugger_case_check_tracer.py', name='thread_func', f_trace='trace_exception'),
            FrameInfo(filename='threading.py', name='run', f_trace='None'),
            FrameInfo(filename='threading.py', name='_bootstrap_inner', f_trace='trace_unhandled_exceptions'),
            FrameInfo(filename='threading.py', name='_bootstrap', f_trace='None'),
            FrameInfo(filename='pydev_monkey.py', name='__call__', f_trace='None')
        ])
    else:
        check_frame_info([
            FrameInfo(filename='_debugger_case_check_tracer.py', name='thread_func', f_trace='trace_exception'),
            FrameInfo(filename='threading.py', name='run', f_trace='None'),
            FrameInfo(filename='threading.py', name='__bootstrap_inner', f_trace='trace_unhandled_exceptions'),
            FrameInfo(filename='threading.py', name='__bootstrap', f_trace='None'),
            FrameInfo(filename='pydev_monkey.py', name='__call__', f_trace='None'),
        ])


th = threading.Thread(target=thread_func)
th.daemon = True
th.start()

event = threading.Event()


def thread_func2():
    try:
        check_frame_info([
            FrameInfo(filename='_debugger_case_check_tracer.py', name='thread_func2', f_trace='trace_exception'),
            FrameInfo(filename='pydev_monkey.py', name='__call__', f_trace='trace_unhandled_exceptions')
        ])
    finally:
        event.set()


start_new_thread(thread_func2, ())

event.wait()
th.join()

# This is a bit tricky: although we waited on the event, there's a slight chance
# that we didn't get the notification because the thread could've stopped executing,
# so, sleep a bit so that the test does not become flaky.
import time
time.sleep(.3)

if sys.version_info[0] >= 3:
    check_frame_info([
        FrameInfo(filename='_debugger_case_check_tracer.py', name='<module>', f_trace='trace_exception'),
        FrameInfo(filename='_pydev_execfile.py', name='execfile', f_trace='None'),
        FrameInfo(filename='pydevd.py', name='_exec', f_trace='trace_unhandled_exceptions'),
        FrameInfo(filename='pydevd.py', name='run', f_trace='trace_dispatch|None'),
        FrameInfo(filename='pydevd.py', name='main', f_trace='trace_dispatch|None'),
        FrameInfo(filename='pydevd.py', name='<module>', f_trace='trace_dispatch|None')
    ])
else:
    check_frame_info([
        FrameInfo(filename='_debugger_case_check_tracer.py', name='<module>', f_trace='trace_exception'),
        FrameInfo(filename='pydevd.py', name='_exec', f_trace='trace_unhandled_exceptions'),
        FrameInfo(filename='pydevd.py', name='run', f_trace='trace_dispatch|None'),
        FrameInfo(filename='pydevd.py', name='main', f_trace='trace_dispatch|None'),
        FrameInfo(filename='pydevd.py', name='<module>', f_trace='trace_dispatch|None'),
    ])

print('TEST SUCEEDED')



# coding: utf-8
import os.path
from _pydevd_bundle.pydevd_constants import IS_WINDOWS
import io
from _pydev_bundle.pydev_log import log_context
import pytest
import sys


@pytest.fixture(autouse=True)
def _reset_ide_os():
    yield
    from pydevd_file_utils import set_ide_os
    set_ide_os('WINDOWS' if sys.platform == 'win32' else 'UNIX')


def test_convert_utilities(tmpdir):
    import pydevd_file_utils

    test_dir = str(tmpdir.mkdir("Test_Convert_Utilities"))

    if IS_WINDOWS:
        normalized = pydevd_file_utils.normcase(test_dir)
        assert isinstance(normalized, str)
        assert normalized.lower() == normalized

        upper_version = os.path.join(test_dir, 'ÁÉÍÓÚ')
        with open(upper_version, 'w') as stream:
            stream.write('test')

        with open(upper_version, 'r') as stream:
            assert stream.read() == 'test'

        with open(pydevd_file_utils.normcase(upper_version), 'r') as stream:
            assert stream.read() == 'test'

        assert '~' not in normalized

        for i in range(3):  # Check if cache is ok.

            if i == 2:
                pydevd_file_utils._listdir_cache.clear()

            assert pydevd_file_utils.get_path_with_real_case('<does not EXIST>') == '<does not EXIST>'
            real_case = pydevd_file_utils.get_path_with_real_case(normalized)
            assert isinstance(real_case, str)
            # Note test_dir itself cannot be compared with because pytest may
            # have passed the case normalized.
            assert real_case.endswith("Test_Convert_Utilities")

            if i == 2:
                # Check that we have the expected paths in the cache.
                assert pydevd_file_utils._listdir_cache[os.path.dirname(normalized).lower()] == ['Test_Convert_Utilities']
                assert pydevd_file_utils._listdir_cache[(os.path.dirname(normalized).lower(), 'Test_Convert_Utilities'.lower())] == real_case

        # Check that it works with a shortened path.
        shortened = pydevd_file_utils.convert_to_short_pathname(normalized)
        assert '~' in shortened
        with_real_case = pydevd_file_utils.get_path_with_real_case(shortened)
        assert with_real_case.endswith('Test_Convert_Utilities')
        assert '~' not in with_real_case

    else:
        # On other platforms, nothing should change
        assert pydevd_file_utils.normcase(test_dir) == test_dir
        assert pydevd_file_utils.get_path_with_real_case(test_dir) == test_dir


def test_source_reference(tmpdir):
    import pydevd_file_utils

    pydevd_file_utils.set_ide_os('WINDOWS')
    if IS_WINDOWS:
        # Client and server are on windows.
        pydevd_file_utils.setup_client_server_paths([('c:\\foo', 'c:\\bar')])

        assert pydevd_file_utils.map_file_to_client('c:\\bar\\my') == ('c:\\foo\\my', True)
        assert pydevd_file_utils.get_client_filename_source_reference('c:\\foo\\my') == 0

        assert pydevd_file_utils.map_file_to_client('c:\\another\\my') == ('c:\\another\\my', False)
        source_reference = pydevd_file_utils.get_client_filename_source_reference('c:\\another\\my')
        assert source_reference != 0
        assert pydevd_file_utils.get_server_filename_from_source_reference(source_reference) == 'c:\\another\\my'

    else:
        # Client on windows and server on unix
        pydevd_file_utils.set_ide_os('WINDOWS')

        pydevd_file_utils.setup_client_server_paths([('c:\\foo', '/bar')])

        assert pydevd_file_utils.map_file_to_client('/bar/my') == ('c:\\foo\\my', True)
        assert pydevd_file_utils.get_client_filename_source_reference('c:\\foo\\my') == 0

        assert pydevd_file_utils.map_file_to_client('/another/my') == ('\\another\\my', False)
        source_reference = pydevd_file_utils.get_client_filename_source_reference('\\another\\my')
        assert source_reference != 0
        assert pydevd_file_utils.get_server_filename_from_source_reference(source_reference) == '/another/my'


@pytest.mark.skipif(sys.platform != 'win32', reason='Windows-only test.')
def test_translate_only_drive():
    import pydevd_file_utils
    assert pydevd_file_utils.get_path_with_real_case('c:\\') == 'C:\\'


def test_to_server_and_to_client(tmpdir):
    try:

        def check(obtained, expected):
            assert obtained == expected, '%s (%s) != %s (%s)' % (obtained, type(obtained), expected, type(expected))
            if isinstance(obtained, tuple):
                assert isinstance(obtained[0], str)
            else:
                assert isinstance(obtained, str)

            if isinstance(expected, tuple):
                assert isinstance(expected[0], str)
            else:
                assert isinstance(expected, str)

        import pydevd_file_utils
        if IS_WINDOWS:
            # Check with made-up files

            pydevd_file_utils.setup_client_server_paths([('c:\\foo', 'c:\\bar'), ('c:\\foo2', 'c:\\bar2')])

            stream = io.StringIO()
            with log_context(0, stream=stream):
                pydevd_file_utils.map_file_to_server('y:\\only_exists_in_client_not_in_server')
            assert r'pydev debugger: unable to find translation for: "y:\only_exists_in_client_not_in_server" in ["c:\foo\", "c:\foo2\", "c:\foo", "c:\foo2"] (please revise your path mappings).' in stream.getvalue()

            # Client and server are on windows.
            pydevd_file_utils.set_ide_os('WINDOWS')
            for in_eclipse, in_python  in ([
                    ('c:\\foo', 'c:\\bar'),
                    ('c:/foo', 'c:\\bar'),
                    ('c:\\foo', 'c:/bar'),
                    ('c:\\foo', 'c:\\bar\\'),
                    ('c:/foo', 'c:\\bar\\'),
                    ('c:\\foo', 'c:/bar/'),
                    ('c:\\foo\\', 'c:\\bar'),
                    ('c:/foo/', 'c:\\bar'),
                    ('c:\\foo\\', 'c:/bar'),

                ]):
                PATHS_FROM_ECLIPSE_TO_PYTHON = [
                    (in_eclipse, in_python)
                ]
                pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
                check(pydevd_file_utils.map_file_to_server('c:\\foo\\my'), 'c:\\bar\\my')
                check(pydevd_file_utils.map_file_to_server('c:/foo/my'), 'c:\\bar\\my')
                check(pydevd_file_utils.map_file_to_server('c:/foo/my/'), 'c:\\bar\\my')
                check(pydevd_file_utils.map_file_to_server('c:\\foo\\áéíóú'.upper()), 'c:\\bar' + '\\áéíóú'.upper())
                check(pydevd_file_utils.map_file_to_client('c:\\bar\\my'), ('c:\\foo\\my', True))

            # Client on unix and server on windows
            pydevd_file_utils.set_ide_os('UNIX')
            for in_eclipse, in_python  in ([
                    ('/foo', 'c:\\bar'),
                    ('/foo', 'c:/bar'),
                    ('/foo', 'c:\\bar\\'),
                    ('/foo', 'c:/bar/'),
                    ('/foo/', 'c:\\bar'),
                    ('/foo/', 'c:\\bar\\'),
                ]):

                PATHS_FROM_ECLIPSE_TO_PYTHON = [
                    (in_eclipse, in_python)
                ]
                pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
                check(pydevd_file_utils.map_file_to_server('/foo/my'), 'c:\\bar\\my')
                check(pydevd_file_utils.map_file_to_client('c:\\bar\\my'), ('/foo/my', True))
                check(pydevd_file_utils.map_file_to_client('c:\\bar\\my\\'), ('/foo/my', True))
                check(pydevd_file_utils.map_file_to_client('c:/bar/my'), ('/foo/my', True))
                check(pydevd_file_utils.map_file_to_client('c:/bar/my/'), ('/foo/my', True))

            # Test with 'real' files
            # Client and server are on windows.
            pydevd_file_utils.set_ide_os('WINDOWS')

            test_dir = pydevd_file_utils.get_path_with_real_case(str(tmpdir.mkdir("Foo")))
            os.makedirs(os.path.join(test_dir, "Another"))

            in_eclipse = os.path.join(os.path.dirname(test_dir), 'Bar')
            in_python = test_dir
            PATHS_FROM_ECLIPSE_TO_PYTHON = [
                (in_eclipse, in_python)
            ]
            pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)

            if pydevd_file_utils.map_file_to_server(in_eclipse) != in_python.lower():
                raise AssertionError('%s != %s\ntmpdir:%s\nin_eclipse: %s\nin_python: %s\ntest_dir: %s' % (
                    pydevd_file_utils.map_file_to_server(in_eclipse), in_python.lower(), tmpdir, in_eclipse, in_python, test_dir))

            found_in_eclipse = pydevd_file_utils.map_file_to_client(in_python)[0]
            assert found_in_eclipse.endswith('Bar')

            assert pydevd_file_utils.map_file_to_server(
                os.path.join(in_eclipse, 'another')) == os.path.join(in_python, 'another').lower()
            found_in_eclipse = pydevd_file_utils.map_file_to_client(
                os.path.join(in_python, 'another'))[0]
            assert found_in_eclipse.endswith('Bar\\Another')

            # Client on unix and server on windows
            pydevd_file_utils.set_ide_os('UNIX')
            in_eclipse = '/foo'
            in_python = test_dir
            PATHS_FROM_ECLIPSE_TO_PYTHON = [
                (in_eclipse, in_python)
            ]
            pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
            assert pydevd_file_utils.map_file_to_server('/foo').lower() == in_python.lower()
            assert pydevd_file_utils.map_file_to_client(in_python) == (in_eclipse, True)

            # Test without translation in place (still needs to fix case and separators)
            pydevd_file_utils.set_ide_os('WINDOWS')
            PATHS_FROM_ECLIPSE_TO_PYTHON = []
            pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
            assert pydevd_file_utils.map_file_to_server(test_dir) == test_dir
            assert pydevd_file_utils.map_file_to_client(test_dir.lower())[0].endswith('\\Foo')
        else:
            # Client on windows and server on unix
            pydevd_file_utils.set_ide_os('WINDOWS')

            PATHS_FROM_ECLIPSE_TO_PYTHON = [
                ('c:\\BAR', '/bar')
            ]

            pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
            assert pydevd_file_utils.map_file_to_server('c:\\bar\\my') == '/bar/my'
            assert pydevd_file_utils.map_file_to_client('/bar/my') == ('c:\\BAR\\my', True)

            for in_eclipse, in_python  in ([
                    ('c:\\foo', '/báéíóúr'),
                    ('c:/foo', '/báéíóúr'),
                    ('c:/foo/', '/báéíóúr'),
                    ('c:/foo/', '/báéíóúr/'),
                    ('c:\\foo\\', '/báéíóúr/'),
                ]):

                PATHS_FROM_ECLIPSE_TO_PYTHON = [
                    (in_eclipse, in_python)
                ]

                pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
                assert pydevd_file_utils.map_file_to_server('c:\\foo\\my') == '/báéíóúr/my'
                assert pydevd_file_utils.map_file_to_server('C:\\foo\\my') == '/báéíóúr/my'
                assert pydevd_file_utils.map_file_to_server('C:\\foo\\MY') == '/báéíóúr/MY'
                assert pydevd_file_utils.map_file_to_server('C:\\foo\\MY\\') == '/báéíóúr/MY'
                assert pydevd_file_utils.map_file_to_server('c:\\foo\\my\\file.py') == '/báéíóúr/my/file.py'
                assert pydevd_file_utils.map_file_to_server('c:\\foo\\my\\other\\file.py') == '/báéíóúr/my/other/file.py'
                assert pydevd_file_utils.map_file_to_server('c:/foo/my') == '/báéíóúr/my'
                assert pydevd_file_utils.map_file_to_server('c:\\foo\\my\\') == '/báéíóúr/my'
                assert pydevd_file_utils.map_file_to_server('c:/foo/my/') == '/báéíóúr/my'

                assert pydevd_file_utils.map_file_to_client('/báéíóúr/my') == ('c:\\foo\\my', True)
                assert pydevd_file_utils.map_file_to_client('/báéíóúr/my/') == ('c:\\foo\\my', True)

                # Files for which there's no translation have only their separators updated.
                assert pydevd_file_utils.map_file_to_client('/usr/bin/x.py') == ('\\usr\\bin\\x.py', False)
                assert pydevd_file_utils.map_file_to_client('/usr/bin') == ('\\usr\\bin', False)
                assert pydevd_file_utils.map_file_to_client('/usr/bin/') == ('\\usr\\bin', False)
                assert pydevd_file_utils.map_file_to_server('\\usr\\bin') == '/usr/bin'
                assert pydevd_file_utils.map_file_to_server('\\usr\\bin\\') == '/usr/bin'

                # When we have a client file and there'd be no translation, and making it absolute would
                # do something as '$cwd/$file_received' (i.e.: $cwd/c:/another in the case below),
                # warn the user that it's not correct and the path that should be translated instead
                # and don't make it absolute.
                assert pydevd_file_utils.map_file_to_server('c:\\Another') == 'c:/Another'

                assert pydevd_file_utils.map_file_to_server('c:/FoO/my/BAR') == '/báéíóúr/my/BAR'
                assert pydevd_file_utils.map_file_to_client('/báéíóúr/my/BAR') == ('c:\\foo\\my\\BAR', True)

            # Client and server on unix
            pydevd_file_utils.set_ide_os('UNIX')
            in_eclipse = '/foo'
            in_python = '/báéíóúr'
            PATHS_FROM_ECLIPSE_TO_PYTHON = [
                (in_eclipse, in_python)
            ]
            pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
            assert pydevd_file_utils.map_file_to_server('/foo/my') == '/báéíóúr/my'
            assert pydevd_file_utils.map_file_to_client('/báéíóúr/my') == ('/foo/my', True)
    finally:
        pydevd_file_utils.setup_client_server_paths([])


def test_relative_paths(tmpdir):
    '''
    We need to check that we can deal with relative paths.

    Use cases:
        - Relative path of file that does not exist:
            Use case is a cython-generated module which is generated from a .pyx which
            is not distributed. In this case we need to resolve the file to a library path file.

        - Relative path of a file that exists but not when resolved from the working directory:
            Use case is a cython-generated module which is generated from a .pyx which is
            distributed. In this case we need to resolve to the real file based on the sys.path
            entries.
    '''
    import pydevd_file_utils
    import sys
    sys.path.append(str(tmpdir))
    try:
        pydevd_file_utils.NORM_PATHS_AND_BASE_CONTAINER.clear()
        pydevd_file_utils.NORM_PATHS_CONTAINER.clear()
        abs_path = pydevd_file_utils.get_abs_path_real_path_and_base_from_file('my_dir/my_file.pyx')[0]
        assert 'site-packages' in abs_path
        assert os.path.normcase(str(tmpdir)) not in abs_path
        assert not pydevd_file_utils.exists('my_dir/my_file.pyx')

        # If the relative file exists when joined with some entry in the PYTHONPATH we'll consider
        # that the relative path points to that absolute path.
        target_dir = os.path.join(str(tmpdir), 'my_dir')
        os.makedirs(target_dir)
        with open(os.path.join(target_dir, 'my_file.pyx'), 'w') as stream:
            stream.write('empty')

        pydevd_file_utils.NORM_PATHS_AND_BASE_CONTAINER.clear()
        pydevd_file_utils.NORM_PATHS_CONTAINER.clear()
        abs_path = pydevd_file_utils.get_abs_path_real_path_and_base_from_file('my_dir/my_file.pyx')[0]
        assert 'site-packages' not in abs_path
        assert str(tmpdir) in abs_path
        assert pydevd_file_utils.exists('my_dir/my_file.pyx')
    finally:
        sys.path.remove(str(tmpdir))


def test_zip_paths(tmpdir):
    import pydevd_file_utils
    import sys
    import zipfile

    for i, zip_basename in enumerate(('MY1.zip', 'my2.egg!')):
        zipfile_path = str(tmpdir.join(zip_basename))
        zip_file = zipfile.ZipFile(zipfile_path, 'w')
        zip_file.writestr('zipped%s/__init__.py' % (i,), '')
        zip_file.writestr('zipped%s/zipped_contents.py' % (i,), 'def call_in_zip():\n    return 1')
        zip_file.close()

        sys.path.append(zipfile_path)
        try:
            import importlib
        except ImportError:
            __import__('zipped%s' % (i,))  # Py2.6 does not have importlib
        else:
            importlib.import_module('zipped%s' % (i,))  # Check that it's importable.

        # Check that we can deal with the zip path.
        assert pydevd_file_utils.exists(zipfile_path)
        abspath, realpath, basename = pydevd_file_utils.get_abs_path_real_path_and_base_from_file(zipfile_path)
        if IS_WINDOWS:
            assert abspath == zipfile_path
            assert basename == zip_basename.lower()
        else:
            assert abspath == zipfile_path
            assert basename == zip_basename

        # Check that we can deal with zip contents.
        for path in [
                zipfile_path + '/zipped%s/__init__.py' % (i,),
                zipfile_path + '/zipped%s/zipped_contents.py' % (i,),
                zipfile_path + '\\zipped%s\\__init__.py' % (i,),
                zipfile_path + '\\zipped%s\\zipped_contents.py' % (i,),
            ]:
            assert pydevd_file_utils.exists(path), 'Expected exists to return True for path:\n%s' % (path,)
            abspath, realpath, basename = pydevd_file_utils.get_abs_path_real_path_and_base_from_file(path)
            assert pydevd_file_utils.exists(abspath), 'Expected exists to return True for path:\n%s' % (abspath,)
            assert pydevd_file_utils.exists(realpath), 'Expected exists to return True for path:\n%s' % (realpath,)

        assert zipfile_path in pydevd_file_utils._ZIP_SEARCH_CACHE, '%s not in %s' % (
            zipfile_path, '\n'.join(sorted(pydevd_file_utils._ZIP_SEARCH_CACHE.keys())))


def test_source_mapping():

    from _pydevd_bundle.pydevd_source_mapping import SourceMapping, SourceMappingEntry
    from _pydevd_bundle import pydevd_api

    class _DummyPyDB(object):
        source_mapping = SourceMapping()
        api_received_breakpoints = {}
        file_to_id_to_line_breakpoint = {}
        file_to_id_to_plugin_breakpoint = {}
        breakpoints = {}

    source_mapping = _DummyPyDB.source_mapping

    mapping = [
        SourceMappingEntry(line=3, end_line=6, runtime_line=5, runtime_source='<cell1>'),
        SourceMappingEntry(line=10, end_line=11, runtime_line=1, runtime_source='<cell2>'),
    ]

    api = pydevd_api.PyDevdAPI()
    py_db = _DummyPyDB()
    filename = 'c:\\temp\\bar.py' if IS_WINDOWS else '/temp/bar.py'
    api.set_source_mapping(py_db, filename, mapping)

    # Map to server
    assert source_mapping.map_to_server(filename, 1) == (filename, 1, False)
    assert source_mapping.map_to_server(filename, 2) == (filename, 2, False)

    assert source_mapping.map_to_server(filename, 3) == ('<cell1>', 5, True)
    assert source_mapping.map_to_server(filename, 4) == ('<cell1>', 6, True)
    assert source_mapping.map_to_server(filename, 5) == ('<cell1>', 7, True)
    assert source_mapping.map_to_server(filename, 6) == ('<cell1>', 8, True)

    assert source_mapping.map_to_server(filename, 7) == (filename, 7, False)

    assert source_mapping.map_to_server(filename, 10) == ('<cell2>', 1, True)
    assert source_mapping.map_to_server(filename, 11) == ('<cell2>', 2, True)

    assert source_mapping.map_to_server(filename, 12) == (filename, 12, False)

    # Map to client
    assert source_mapping.map_to_client(filename, 1) == (filename, 1, False)
    assert source_mapping.map_to_client(filename, 2) == (filename, 2, False)

    assert source_mapping.map_to_client('<cell1>', 5) == (filename, 3, True)
    assert source_mapping.map_to_client('<cell1>', 6) == (filename, 4, True)
    assert source_mapping.map_to_client('<cell1>', 7) == (filename, 5, True)
    assert source_mapping.map_to_client('<cell1>', 8) == (filename, 6, True)

    assert source_mapping.map_to_client(filename, 7) == (filename, 7, False)

    assert source_mapping.map_to_client('<cell2>', 1) == (filename, 10, True)
    assert source_mapping.map_to_client('<cell2>', 2) == (filename, 11, True)

    assert source_mapping.map_to_client(filename, 12) == (filename, 12, False)


@pytest.mark.skipif(IS_WINDOWS, reason='Linux-only test')
def test_mapping_conflict_to_client():
    import pydevd_file_utils

    path_mappings = []
    for pathMapping in _MAPPING_CONFLICT:
        localRoot = pathMapping.get('localRoot', '')
        remoteRoot = pathMapping.get('remoteRoot', '')
        if (localRoot != '') and (remoteRoot != ''):
            path_mappings.append((localRoot, remoteRoot))

    pydevd_file_utils.setup_client_server_paths(path_mappings)

    assert pydevd_file_utils.map_file_to_client('/opt/pathsomething/foo.py') == \
        ('/var/home/p2/foo.py', True)

    assert pydevd_file_utils.map_file_to_client('/opt/v2/pathsomething/foo.py') == \
        ('/var/home/p4/foo.py', True)

    # This is an odd case, but the user didn't really put a slash in the end,
    # so, it's possible that this is what the user actually wants.
    assert pydevd_file_utils.map_file_to_client('/opt/v2/path_r1/foo.py') == \
        ('/var/home/p3_r1/foo.py', True)

    # The client said both local and remote end with a slash, so, we can only
    # match it with the slash in the end.
    assert pydevd_file_utils.map_file_to_client('/opt/pathsomething_foo.py') == \
        ('/opt/pathsomething_foo.py', False)


_MAPPING_CONFLICT = [
    {
        "localRoot": "/var/home/p1/",
        "remoteRoot": "/opt/path/"
    },
    {
        "localRoot": "/var/home/p2/",
        "remoteRoot": "/opt/pathsomething/"
    },
    {
        "localRoot": "/var/home/p3",
        "remoteRoot": "/opt/v2/path"
    },
    {
        "localRoot": "/var/home/p4",
        "remoteRoot": "/opt/v2/pathsomething"
    },
]


@pytest.mark.skipif(IS_WINDOWS, reason='Linux-only test')
def test_mapping_conflict_to_server():
    import pydevd_file_utils

    path_mappings = []
    for pathMapping in _MAPPING_CONFLICT_TO_SERVER:
        localRoot = pathMapping.get('localRoot', '')
        remoteRoot = pathMapping.get('remoteRoot', '')
        if (localRoot != '') and (remoteRoot != ''):
            path_mappings.append((localRoot, remoteRoot))

    pydevd_file_utils.setup_client_server_paths(path_mappings)

    assert pydevd_file_utils.map_file_to_server('/opt/pathsomething/foo.py') == '/var/home/p2/foo.py'

    assert pydevd_file_utils.map_file_to_server('/opt/v2/pathsomething/foo.py') == '/var/home/p4/foo.py'

    # This is an odd case, but the user didn't really put a slash in the end,
    # so, it's possible that this is what the user actually wants.
    assert pydevd_file_utils.map_file_to_server('/opt/v2/path_r1/foo.py') == '/var/home/p3_r1/foo.py'

    # The client said both local and remote end with a slash, so, we can only
    # match it with the slash in the end.
    assert pydevd_file_utils.map_file_to_server('/opt/pathsomething_foo.py') == '/opt/pathsomething_foo.py'


_MAPPING_CONFLICT_TO_SERVER = [
    {
        "remoteRoot": "/var/home/p1/",
        "localRoot": "/opt/path/"
    },
    {
        "remoteRoot": "/var/home/p2/",
        "localRoot": "/opt/pathsomething/"
    },
    {
        "remoteRoot": "/var/home/p3",
        "localRoot": "/opt/v2/path"
    },
    {
        "remoteRoot": "/var/home/p4",
        "localRoot": "/opt/v2/pathsomething"
    },
]




import sys
from _pydevd_bundle.pydevd_constants import int_types, GENERATED_LEN_ATTR_NAME
from _pydevd_bundle.pydevd_resolver import MAX_ITEMS_TO_HANDLE, TOO_LARGE_ATTR
from _pydevd_bundle import pydevd_frame_utils


def get_frame():
    var1 = 1
    var2 = [var1]
    var3 = {33: [var1]}
    return sys._getframe()


def check_vars_dict_expected(as_dict, expected):
    assert as_dict == expected


class _DummyPyDB(object):

    def __init__(self):
        from _pydevd_bundle.pydevd_api import PyDevdAPI
        self.variable_presentation = PyDevdAPI.VariablePresentation()


def test_suspended_frames_manager():
    from _pydevd_bundle.pydevd_suspended_frames import SuspendedFramesManager
    from _pydevd_bundle.pydevd_utils import DAPGrouper
    suspended_frames_manager = SuspendedFramesManager()
    py_db = _DummyPyDB()
    with suspended_frames_manager.track_frames(py_db) as tracker:
        # : :type tracker: _FramesTracker
        thread_id = 'thread1'
        frame = get_frame()
        tracker.track(thread_id, pydevd_frame_utils.create_frames_list_from_frame(frame))

        assert suspended_frames_manager.get_thread_id_for_variable_reference(id(frame)) == thread_id

        variable = suspended_frames_manager.get_variable(id(frame))

        # Should be properly sorted.
        assert ['var1', 'var2', 'var3'] == [x.get_name()for x in variable.get_children_variables()]

        as_dict = dict((x.get_name(), x.get_var_data()) for x in variable.get_children_variables())
        var_reference = as_dict['var2'].pop('variablesReference')
        assert isinstance(var_reference, int_types)  # The variable reference is always a new int.
        assert isinstance(as_dict['var3'].pop('variablesReference'), int_types)  # The variable reference is always a new int.

        check_vars_dict_expected(as_dict, {
            'var1': {'name': 'var1', 'value': '1', 'type': 'int', 'evaluateName': 'var1', 'variablesReference': 0},
            'var2': {'name': 'var2', 'value': '[1]', 'type': 'list', 'evaluateName': 'var2'},
            'var3': {'name': 'var3', 'value': '{33: [1]}', 'type': 'dict', 'evaluateName': 'var3'}
        })

        # Now, same thing with a different format.
        as_dict = dict((x.get_name(), x.get_var_data(fmt={'hex': True})) for x in variable.get_children_variables())
        var_reference = as_dict['var2'].pop('variablesReference')
        assert isinstance(var_reference, int_types)  # The variable reference is always a new int.
        assert isinstance(as_dict['var3'].pop('variablesReference'), int_types)  # The variable reference is always a new int.

        check_vars_dict_expected(as_dict, {
            'var1': {'name': 'var1', 'value': '0x1', 'type': 'int', 'evaluateName': 'var1', 'variablesReference': 0},
            'var2': {'name': 'var2', 'value': '[0x1]', 'type': 'list', 'evaluateName': 'var2'},
            'var3': {'name': 'var3', 'value': '{0x21: [0x1]}', 'type': 'dict', 'evaluateName': 'var3'}
        })

        var2 = dict((x.get_name(), x) for x in variable.get_children_variables())['var2']
        children_vars = var2.get_children_variables()
        as_dict = (dict([x.get_name(), x.get_var_data()] for x in children_vars if x.get_name() not in DAPGrouper.SCOPES_SORTED))
        assert as_dict == {
            '0': {'name': '0', 'value': '1', 'type': 'int', 'evaluateName': 'var2[0]', 'variablesReference': 0 },
            GENERATED_LEN_ATTR_NAME: {'name': GENERATED_LEN_ATTR_NAME, 'value': '1', 'type': 'int', 'evaluateName': 'len(var2)', 'variablesReference': 0, 'presentationHint': {'attributes': ['readOnly']}, },
        }

        var3 = dict((x.get_name(), x) for x in variable.get_children_variables())['var3']
        children_vars = var3.get_children_variables()
        as_dict = (dict([x.get_name(), x.get_var_data()] for x in children_vars if x.get_name() not in DAPGrouper.SCOPES_SORTED))
        assert isinstance(as_dict['33'].pop('variablesReference'), int_types)  # The variable reference is always a new int.

        check_vars_dict_expected(as_dict, {
            '33': {'name': '33', 'value': "[1]", 'type': 'list', 'evaluateName': 'var3[33]'},
            GENERATED_LEN_ATTR_NAME: {'name': GENERATED_LEN_ATTR_NAME, 'value': '1', 'type': 'int', 'evaluateName': 'len(var3)', 'variablesReference': 0, 'presentationHint': {'attributes': ['readOnly']}, }
        })


_NUMBER_OF_ITEMS_TO_CREATE = MAX_ITEMS_TO_HANDLE + 300


def get_dict_large_frame():
    obj = {}
    for idx in range(_NUMBER_OF_ITEMS_TO_CREATE):
        obj[idx] = (1)
    return sys._getframe()


def get_set_large_frame():
    obj = set()
    for idx in range(_NUMBER_OF_ITEMS_TO_CREATE):
        obj.add(idx)
    return sys._getframe()


def get_tuple_large_frame():
    obj = tuple(range(_NUMBER_OF_ITEMS_TO_CREATE))
    return sys._getframe()


def test_get_child_variables():
    from _pydevd_bundle.pydevd_suspended_frames import SuspendedFramesManager
    suspended_frames_manager = SuspendedFramesManager()
    py_db = _DummyPyDB()
    for frame in (
        get_dict_large_frame(),
        get_set_large_frame(),
        get_tuple_large_frame(),
        ):
        with suspended_frames_manager.track_frames(py_db) as tracker:
            # : :type tracker: _FramesTracker
            thread_id = 'thread1'
            tracker.track(thread_id, pydevd_frame_utils.create_frames_list_from_frame(frame))

            assert suspended_frames_manager.get_thread_id_for_variable_reference(id(frame)) == thread_id

            variable = suspended_frames_manager.get_variable(id(frame))

            children_variables = variable.get_child_variable_named('obj').get_children_variables()
            assert len(children_variables) < _NUMBER_OF_ITEMS_TO_CREATE

            found_too_large = False
            found_len = False
            for x in children_variables:
                if x.name == TOO_LARGE_ATTR:
                    var_data = x.get_var_data()
                    assert 'readOnly' in var_data['presentationHint']['attributes']
                    found_too_large = True
                elif x.name == GENERATED_LEN_ATTR_NAME:
                    found_len = True

            if not found_too_large:
                raise AssertionError('Expected to find variable named: %s' % (TOO_LARGE_ATTR,))
            if not found_len:
                raise AssertionError('Expected to find variable named: len()')




def step_over():
    print("don't step here")

def step_into():
    print('step into')

def step_out():
    print('step out')  # Break here 1
    print("don't step here")

def Call():
    step_over()  # Break here 2
    step_into()
    step_out()
    print('stepped out')

if __name__ == '__main__':
    Call()
    print('TEST SUCEEDED!')


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# winapi.py: Windows API-Python interface (removes dependency on pywin32)
#
# Copyright (C) 2007 Thomas Heller <theller@ctypes.org>
# Copyright (C) 2010 Will McGugan <will@willmcgugan.com>
# Copyright (C) 2010 Ryan Kelly <ryan@rfk.id.au>
# Copyright (C) 2010 Yesudeep Mangalapilly <yesudeep@gmail.com>
# Copyright (C) 2014 Thomas Amland
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and / or other materials provided with the distribution.
# * Neither the name of the organization nor the names of its contributors may
#   be used to endorse or promote products derived from this software without
#   specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Portions of this code were taken from pyfilesystem, which uses the above
# new BSD license.

import ctypes.wintypes
from functools import reduce

LPVOID = ctypes.wintypes.LPVOID

# Invalid handle value.
INVALID_HANDLE_VALUE = ctypes.c_void_p(-1).value

# File notification constants.
FILE_NOTIFY_CHANGE_FILE_NAME = 0x01
FILE_NOTIFY_CHANGE_DIR_NAME = 0x02
FILE_NOTIFY_CHANGE_ATTRIBUTES = 0x04
FILE_NOTIFY_CHANGE_SIZE = 0x08
FILE_NOTIFY_CHANGE_LAST_WRITE = 0x010
FILE_NOTIFY_CHANGE_LAST_ACCESS = 0x020
FILE_NOTIFY_CHANGE_CREATION = 0x040
FILE_NOTIFY_CHANGE_SECURITY = 0x0100

FILE_FLAG_BACKUP_SEMANTICS = 0x02000000
FILE_FLAG_OVERLAPPED = 0x40000000
FILE_LIST_DIRECTORY = 1
FILE_SHARE_READ = 0x01
FILE_SHARE_WRITE = 0x02
FILE_SHARE_DELETE = 0x04
OPEN_EXISTING = 3

VOLUME_NAME_NT = 0x02

# File action constants.
FILE_ACTION_CREATED = 1
FILE_ACTION_DELETED = 2
FILE_ACTION_MODIFIED = 3
FILE_ACTION_RENAMED_OLD_NAME = 4
FILE_ACTION_RENAMED_NEW_NAME = 5
FILE_ACTION_DELETED_SELF = 0xFFFE
FILE_ACTION_OVERFLOW = 0xFFFF

# Aliases
FILE_ACTION_ADDED = FILE_ACTION_CREATED
FILE_ACTION_REMOVED = FILE_ACTION_DELETED
FILE_ACTION_REMOVED_SELF = FILE_ACTION_DELETED_SELF

THREAD_TERMINATE = 0x0001

# IO waiting constants.
WAIT_ABANDONED = 0x00000080
WAIT_IO_COMPLETION = 0x000000C0
WAIT_OBJECT_0 = 0x00000000
WAIT_TIMEOUT = 0x00000102

# Error codes
ERROR_OPERATION_ABORTED = 995


class OVERLAPPED(ctypes.Structure):
    _fields_ = [('Internal', LPVOID),
                ('InternalHigh', LPVOID),
                ('Offset', ctypes.wintypes.DWORD),
                ('OffsetHigh', ctypes.wintypes.DWORD),
                ('Pointer', LPVOID),
                ('hEvent', ctypes.wintypes.HANDLE),
                ]


def _errcheck_bool(value, func, args):
    if not value:
        raise ctypes.WinError()
    return args


def _errcheck_handle(value, func, args):
    if not value:
        raise ctypes.WinError()
    if value == INVALID_HANDLE_VALUE:
        raise ctypes.WinError()
    return args


def _errcheck_dword(value, func, args):
    if value == 0xFFFFFFFF:
        raise ctypes.WinError()
    return args


kernel32 = ctypes.WinDLL("kernel32")

ReadDirectoryChangesW = kernel32.ReadDirectoryChangesW
ReadDirectoryChangesW.restype = ctypes.wintypes.BOOL
ReadDirectoryChangesW.errcheck = _errcheck_bool
ReadDirectoryChangesW.argtypes = (
    ctypes.wintypes.HANDLE,  # hDirectory
    LPVOID,  # lpBuffer
    ctypes.wintypes.DWORD,  # nBufferLength
    ctypes.wintypes.BOOL,  # bWatchSubtree
    ctypes.wintypes.DWORD,  # dwNotifyFilter
    ctypes.POINTER(ctypes.wintypes.DWORD),  # lpBytesReturned
    ctypes.POINTER(OVERLAPPED),  # lpOverlapped
    LPVOID  # FileIOCompletionRoutine # lpCompletionRoutine
)

CreateFileW = kernel32.CreateFileW
CreateFileW.restype = ctypes.wintypes.HANDLE
CreateFileW.errcheck = _errcheck_handle
CreateFileW.argtypes = (
    ctypes.wintypes.LPCWSTR,  # lpFileName
    ctypes.wintypes.DWORD,  # dwDesiredAccess
    ctypes.wintypes.DWORD,  # dwShareMode
    LPVOID,  # lpSecurityAttributes
    ctypes.wintypes.DWORD,  # dwCreationDisposition
    ctypes.wintypes.DWORD,  # dwFlagsAndAttributes
    ctypes.wintypes.HANDLE  # hTemplateFile
)

CloseHandle = kernel32.CloseHandle
CloseHandle.restype = ctypes.wintypes.BOOL
CloseHandle.argtypes = (
    ctypes.wintypes.HANDLE,  # hObject
)

CancelIoEx = kernel32.CancelIoEx
CancelIoEx.restype = ctypes.wintypes.BOOL
CancelIoEx.errcheck = _errcheck_bool
CancelIoEx.argtypes = (
    ctypes.wintypes.HANDLE,  # hObject
    ctypes.POINTER(OVERLAPPED)  # lpOverlapped
)

CreateEvent = kernel32.CreateEventW
CreateEvent.restype = ctypes.wintypes.HANDLE
CreateEvent.errcheck = _errcheck_handle
CreateEvent.argtypes = (
    LPVOID,  # lpEventAttributes
    ctypes.wintypes.BOOL,  # bManualReset
    ctypes.wintypes.BOOL,  # bInitialState
    ctypes.wintypes.LPCWSTR,  # lpName
)

SetEvent = kernel32.SetEvent
SetEvent.restype = ctypes.wintypes.BOOL
SetEvent.errcheck = _errcheck_bool
SetEvent.argtypes = (
    ctypes.wintypes.HANDLE,  # hEvent
)

WaitForSingleObjectEx = kernel32.WaitForSingleObjectEx
WaitForSingleObjectEx.restype = ctypes.wintypes.DWORD
WaitForSingleObjectEx.errcheck = _errcheck_dword
WaitForSingleObjectEx.argtypes = (
    ctypes.wintypes.HANDLE,  # hObject
    ctypes.wintypes.DWORD,  # dwMilliseconds
    ctypes.wintypes.BOOL,  # bAlertable
)

CreateIoCompletionPort = kernel32.CreateIoCompletionPort
CreateIoCompletionPort.restype = ctypes.wintypes.HANDLE
CreateIoCompletionPort.errcheck = _errcheck_handle
CreateIoCompletionPort.argtypes = (
    ctypes.wintypes.HANDLE,  # FileHandle
    ctypes.wintypes.HANDLE,  # ExistingCompletionPort
    LPVOID,  # CompletionKey
    ctypes.wintypes.DWORD,  # NumberOfConcurrentThreads
)

GetQueuedCompletionStatus = kernel32.GetQueuedCompletionStatus
GetQueuedCompletionStatus.restype = ctypes.wintypes.BOOL
GetQueuedCompletionStatus.errcheck = _errcheck_bool
GetQueuedCompletionStatus.argtypes = (
    ctypes.wintypes.HANDLE,  # CompletionPort
    LPVOID,  # lpNumberOfBytesTransferred
    LPVOID,  # lpCompletionKey
    ctypes.POINTER(OVERLAPPED),  # lpOverlapped
    ctypes.wintypes.DWORD,  # dwMilliseconds
)

PostQueuedCompletionStatus = kernel32.PostQueuedCompletionStatus
PostQueuedCompletionStatus.restype = ctypes.wintypes.BOOL
PostQueuedCompletionStatus.errcheck = _errcheck_bool
PostQueuedCompletionStatus.argtypes = (
    ctypes.wintypes.HANDLE,  # CompletionPort
    ctypes.wintypes.DWORD,  # lpNumberOfBytesTransferred
    ctypes.wintypes.DWORD,  # lpCompletionKey
    ctypes.POINTER(OVERLAPPED),  # lpOverlapped
)


GetFinalPathNameByHandleW = kernel32.GetFinalPathNameByHandleW
GetFinalPathNameByHandleW.restype = ctypes.wintypes.DWORD
GetFinalPathNameByHandleW.errcheck = _errcheck_dword
GetFinalPathNameByHandleW.argtypes = (
    ctypes.wintypes.HANDLE,  # hFile
    ctypes.wintypes.LPWSTR,  # lpszFilePath
    ctypes.wintypes.DWORD,  # cchFilePath
    ctypes.wintypes.DWORD,  # DWORD
)


class FILE_NOTIFY_INFORMATION(ctypes.Structure):
    _fields_ = [("NextEntryOffset", ctypes.wintypes.DWORD),
                ("Action", ctypes.wintypes.DWORD),
                ("FileNameLength", ctypes.wintypes.DWORD),
                # ("FileName", (ctypes.wintypes.WCHAR * 1))]
                ("FileName", (ctypes.c_char * 1))]


LPFNI = ctypes.POINTER(FILE_NOTIFY_INFORMATION)


# We don't need to recalculate these flags every time a call is made to
# the win32 API functions.
WATCHDOG_FILE_FLAGS = FILE_FLAG_BACKUP_SEMANTICS
WATCHDOG_FILE_SHARE_FLAGS = reduce(
    lambda x, y: x | y, [
        FILE_SHARE_READ,
        FILE_SHARE_WRITE,
        FILE_SHARE_DELETE,
    ])
WATCHDOG_FILE_NOTIFY_FLAGS = reduce(
    lambda x, y: x | y, [
        FILE_NOTIFY_CHANGE_FILE_NAME,
        FILE_NOTIFY_CHANGE_DIR_NAME,
        FILE_NOTIFY_CHANGE_ATTRIBUTES,
        FILE_NOTIFY_CHANGE_SIZE,
        FILE_NOTIFY_CHANGE_LAST_WRITE,
        FILE_NOTIFY_CHANGE_SECURITY,
        FILE_NOTIFY_CHANGE_LAST_ACCESS,
        FILE_NOTIFY_CHANGE_CREATION,
    ])

BUFFER_SIZE = 2048


def _parse_event_buffer(readBuffer, nBytes):
    results = []
    while nBytes > 0:
        fni = ctypes.cast(readBuffer, LPFNI)[0]
        ptr = ctypes.addressof(fni) + FILE_NOTIFY_INFORMATION.FileName.offset
        # filename = ctypes.wstring_at(ptr, fni.FileNameLength)
        filename = ctypes.string_at(ptr, fni.FileNameLength)
        results.append((fni.Action, filename.decode('utf-16')))
        numToSkip = fni.NextEntryOffset
        if numToSkip <= 0:
            break
        readBuffer = readBuffer[numToSkip:]
        nBytes -= numToSkip  # numToSkip is long. nBytes should be long too.
    return results


def _is_observed_path_deleted(handle, path):
    # Comparison of observed path and actual path, returned by
    # GetFinalPathNameByHandleW. If directory moved to the trash bin, or
    # deleted, actual path will not be equal to observed path.
    buff = ctypes.create_unicode_buffer(BUFFER_SIZE)
    GetFinalPathNameByHandleW(handle, buff, BUFFER_SIZE, VOLUME_NAME_NT)
    return buff.value != path


def _generate_observed_path_deleted_event():
    # Create synthetic event for notify that observed directory is deleted
    path = ctypes.create_unicode_buffer('.')
    event = FILE_NOTIFY_INFORMATION(0, FILE_ACTION_DELETED_SELF, len(path), path.value.encode("utf-8"))
    event_size = ctypes.sizeof(event)
    buff = ctypes.create_string_buffer(BUFFER_SIZE)
    ctypes.memmove(buff, ctypes.addressof(event), event_size)
    return buff, event_size


def get_directory_handle(path):
    """Returns a Windows handle to the specified directory path."""
    return CreateFileW(path, FILE_LIST_DIRECTORY, WATCHDOG_FILE_SHARE_FLAGS,
                       None, OPEN_EXISTING, WATCHDOG_FILE_FLAGS, None)


def close_directory_handle(handle):
    try:
        CancelIoEx(handle, None)  # force ReadDirectoryChangesW to return
        CloseHandle(handle)       # close directory handle
    except WindowsError:
        try:
            CloseHandle(handle)   # close directory handle
        except Exception:
            return


def read_directory_changes(handle, path, recursive):
    """Read changes to the directory using the specified directory handle.

    http://timgolden.me.uk/pywin32-docs/win32file__ReadDirectoryChangesW_meth.html
    """
    event_buffer = ctypes.create_string_buffer(BUFFER_SIZE)
    nbytes = ctypes.wintypes.DWORD()
    try:
        ReadDirectoryChangesW(handle, ctypes.byref(event_buffer),
                              len(event_buffer), recursive,
                              WATCHDOG_FILE_NOTIFY_FLAGS,
                              ctypes.byref(nbytes), None, None)
    except WindowsError as e:
        if e.winerror == ERROR_OPERATION_ABORTED:
            return [], 0

        # Handle the case when the root path is deleted
        if _is_observed_path_deleted(handle, path):
            return _generate_observed_path_deleted_event()

        raise e

    # Python 2/3 compat
    try:
        int_class = long
    except NameError:
        int_class = int
    return event_buffer.raw, int_class(nbytes.value)


class WinAPINativeEvent(object):
    def __init__(self, action, src_path):
        self.action = action
        self.src_path = src_path

    @property
    def is_added(self):
        return self.action == FILE_ACTION_CREATED

    @property
    def is_removed(self):
        return self.action == FILE_ACTION_REMOVED

    @property
    def is_modified(self):
        return self.action == FILE_ACTION_MODIFIED

    @property
    def is_renamed_old(self):
        return self.action == FILE_ACTION_RENAMED_OLD_NAME

    @property
    def is_renamed_new(self):
        return self.action == FILE_ACTION_RENAMED_NEW_NAME

    @property
    def is_removed_self(self):
        return self.action == FILE_ACTION_REMOVED_SELF

    def __repr__(self):
        return ("<%s: action=%d, src_path=%r>" % (
                type(self).__name__, self.action, self.src_path))


def read_events(handle, path, recursive):
    buf, nbytes = read_directory_changes(handle, path, recursive)
    events = _parse_event_buffer(buf, nbytes)
    return [WinAPINativeEvent(action, src_path) for action, src_path in events]



from robocorp_ls_core.lsp import CompletionItemTypedDict, CompletionItemKind
from typing import Any, List, Optional
from robocorp_ls_core.protocols import TypedDict


class MonacoCompletionItemKind:
    Method = 0
    Function = 1
    Constructor = 2
    Field = 3
    Variable = 4
    Class = 5
    Struct = 6
    Interface = 7
    Module = 8
    Property = 9
    Event = 10
    Operator = 11
    Unit = 12
    Value = 13
    Constant = 14
    Enum = 15
    EnumMember = 16
    Keyword = 17
    Text = 18
    Color = 19
    File = 20
    Reference = 21
    Customcolor = 22
    Folder = 23
    TypeParameter = 24
    User = 25
    Issue = 26
    Snippet = 27


_lsp_completion_item_kind_to_monaco = {
    CompletionItemKind.Method: MonacoCompletionItemKind.Method,
    CompletionItemKind.Function: MonacoCompletionItemKind.Function,
    CompletionItemKind.Constructor: MonacoCompletionItemKind.Constructor,
    CompletionItemKind.Field: MonacoCompletionItemKind.Field,
    CompletionItemKind.Variable: MonacoCompletionItemKind.Variable,
    CompletionItemKind.Class: MonacoCompletionItemKind.Class,
    CompletionItemKind.Interface: MonacoCompletionItemKind.Interface,
    CompletionItemKind.Struct: MonacoCompletionItemKind.Struct,
    CompletionItemKind.Module: MonacoCompletionItemKind.Module,
    CompletionItemKind.Property: MonacoCompletionItemKind.Property,
    CompletionItemKind.Unit: MonacoCompletionItemKind.Unit,
    CompletionItemKind.Value: MonacoCompletionItemKind.Value,
    CompletionItemKind.Constant: MonacoCompletionItemKind.Constant,
    CompletionItemKind.Enum: MonacoCompletionItemKind.Enum,
    CompletionItemKind.EnumMember: MonacoCompletionItemKind.EnumMember,
    CompletionItemKind.Keyword: MonacoCompletionItemKind.Keyword,
    CompletionItemKind.Snippet: MonacoCompletionItemKind.Snippet,
    CompletionItemKind.Text: MonacoCompletionItemKind.Text,
    CompletionItemKind.Color: MonacoCompletionItemKind.Color,
    CompletionItemKind.File: MonacoCompletionItemKind.File,
    CompletionItemKind.Reference: MonacoCompletionItemKind.Reference,
    CompletionItemKind.Folder: MonacoCompletionItemKind.Folder,
    CompletionItemKind.Event: MonacoCompletionItemKind.Event,
    CompletionItemKind.Operator: MonacoCompletionItemKind.Operator,
    CompletionItemKind.TypeParameter: MonacoCompletionItemKind.TypeParameter,
    CompletionItemKind.User: MonacoCompletionItemKind.User,
    CompletionItemKind.Issue: MonacoCompletionItemKind.Issue,
}


class MonacoCommandTypedDict(TypedDict, total=False):
    id: str
    title: str
    tooltip: Optional[str]
    arguments: Optional[list]


class MonacoCompletionItemTypedDict(TypedDict, total=False):
    #
    # The label of this completion item. By default
    # this is also the text that is inserted when selecting
    # this completion.
    #
    label: str  # string | CompletionItemLabel
    #
    # The kind of this completion item. Based on the kind
    # an icon is chosen by the editor.
    #
    kind: int  # MonacoCompletionItemKind
    #
    # A modifier to the `kind` which affect how the item
    # is rendered, e.g. Deprecated is rendered with a strikeout
    #
    tags: Optional[List[Any]]  # ReadonlyArray<CompletionItemTag>
    #
    # A human-readable string with additional information
    # about this item, like type or symbol information.
    #
    detail: Optional[str]
    #
    # A human-readable string that represents a doc-comment.
    #
    documentation: Optional[str]  # string | IMarkdownString
    #
    # A string that should be used when comparing this item
    # with other items. When `falsy` the [label](#CompletionItem.label)
    # is used.
    #
    sortText: Optional[str]
    #
    # A string that should be used when filtering a set of
    # completion items. When `falsy` the [label](#CompletionItem.label)
    # is used.
    #
    filterText: Optional[str]
    #
    # Select this item when showing. *Note* that only one completion item can be selected and
    # that the editor decides which item that is. The rule is that the *first* item of those
    # that match best is selected.
    #
    preselect: Optional[bool]
    #
    # A string or snippet that should be inserted in a document when selecting
    # this completion.
    # is used.
    #
    insertText: str
    #
    # Addition rules (as bitmask) that should be applied when inserting
    # this completion.
    #
    insertTextRules: Optional[Any]  # CompletionItemInsertTextRule
    #
    # A range of text that should be replaced by this completion item.
    #
    # Defaults to a range from the start of the [current word](#TextDocument.getWordRangeAtPosition) to the
    # current position.
    #
    # *Note:* The range must be a [single line](#Range.isSingleLine) and it must
    # [contain](#Range.contains) the position at which completion has been [requested](#CompletionItemProvider.provideCompletionItems).
    #
    range: Optional[Any]  # IRange | {insert: IRange; replace: IRange}
    #
    # An optional set of characters that when pressed while this completion is active will accept it first and
    # then type that character. *Note* that all commit characters should have `length=1` and that superfluous
    # characters will be ignored.
    #
    commitCharacters: Optional[List[str]]
    #
    # An optional array of additional text edits that are applied when
    # selecting this completion. Edits must not overlap with the main edit
    # nor with themselves.
    #
    additionalTextEdits: Optional[Any]  # editor.ISingleEditOperation[]
    #
    # A command that should be run upon acceptance of this item.
    #
    command: Optional[MonacoCommandTypedDict]

    # Not really in the official docs, but if present it'll be persisted
    # so that it's possible to resolve a completion item.
    data: Any


class CompletionItemInsertTextRule:
    #
    # Adjust whitespace/indentation of multiline insert texts to
    # match the current line indentation.
    # /
    KeepWhitespace = 1
    #
    # `insertText` is a snippet.
    InsertAsSnippet = 4


def convert_to_monaco_completion(
    lsp_completion: CompletionItemTypedDict, line_delta: int, col_delta: int, uri: str
) -> MonacoCompletionItemTypedDict:
    """
    Completions from monaco are different from the completions in the language
    server, so, we need to convert from one to the other.
    """
    from robocorp_ls_core.lsp import InsertTextFormat

    documentation = lsp_completion.get("documentation")
    if isinstance(documentation, dict):  # MarkupContent
        # Should render if markdown...
        documentation = documentation.get("value", "")

    insert_text_format = lsp_completion.get(
        "insertTextFormat", InsertTextFormat.PlainText
    )
    insert_text_rules = 0
    if insert_text_format == InsertTextFormat.Snippet:
        insert_text_rules |= CompletionItemInsertTextRule.InsertAsSnippet

    if lsp_completion.get("insertTextMode") == 2:
        insert_text_rules |= CompletionItemInsertTextRule.KeepWhitespace

    label: str = lsp_completion["label"]
    insert_text: str = lsp_completion.get("insertText", label)  # type: ignore

    text_edit = lsp_completion.get("textEdit")
    range_ = None
    if text_edit:
        insert_text = text_edit["newText"]
        range_ = text_edit["range"]
        if range_:
            # Note: must convert from 0-based to 1-based
            range_["startLineNumber"] = range_["start"]["line"] - line_delta + 1
            range_["startColumn"] = range_["start"]["character"] - col_delta + 1
            range_["endLineNumber"] = range_["end"]["line"] - line_delta + 1
            range_["endColumn"] = range_["end"]["character"] - col_delta + 1

    # Note: we don't cover all, just the things we know are required for the robot lsp usage.
    ret: MonacoCompletionItemTypedDict = {
        "label": label,
        "kind": _lsp_completion_item_kind_to_monaco.get(lsp_completion["kind"], 0),
        "insertText": insert_text,
        "insertTextRules": insert_text_rules,
    }
    if documentation is not None:
        ret["documentation"] = documentation

    if range_ is not None:
        ret["range"] = range_

    sort_text = lsp_completion.get("sortText")
    if sort_text is not None:
        ret["sortText"] = sort_text

    filter_text = lsp_completion.get("filterText")
    if filter_text is not None:
        ret["filterText"] = filter_text

    preselect = lsp_completion.get("preselect")
    if preselect is not None:
        ret["preselect"] = preselect

    detail = lsp_completion.get("detail")
    if detail is not None:
        ret["detail"] = detail

    data = lsp_completion.get("data")
    if data is not None:
        ret["data"] = data

    additional_text_edits = lsp_completion.get("additionalTextEdits")
    if additional_text_edits is not None:
        # Ok, it seems we have some auto-import. Changing the console contents
        # is actually pretty tricky if the user was editing a task as it'd need
        # to actually create a *** Settings *** and a *** Tasks *** and reindent.
        # all the contents (it may not actually be all that difficult, but then
        # the final result to the user may be a bit surprising).
        # So, instead of doing that, a command is issued to add the
        # related import directly to the console.
        if additional_text_edits and len(additional_text_edits) == 1:
            text_edit = additional_text_edits[0]
            command: MonacoCommandTypedDict = {
                "title": "Interactive Console",
                "id": "robot.completion.additionalTextEdit",
                "arguments": [{"code": text_edit["newText"], "uri": uri}],
            }
            ret["command"] = command
    return ret



import inspect
import sys
import unittest

from _pydevd_bundle.pydevd_save_locals import save_locals
from _pydevd_bundle.pydevd_constants import IS_JYTHON, IS_IRONPYTHON
import pytest


def use_save_locals(name, value):
    """
    Attempt to set the local of the given name to value, using locals_to_fast.
    """
    frame = inspect.currentframe().f_back
    locals_dict = frame.f_locals
    locals_dict[name] = value

    save_locals(frame)


def check_method(fn):
    """
    A harness for testing methods that attempt to modify the values of locals on the stack.
    """
    x = 1

    # The method 'fn' should attempt to set x = 2 in the current frame.
    fn('x', 2)

    return x



@pytest.mark.skipif(IS_JYTHON or IS_IRONPYTHON, reason='CPython/pypy only')
class TestSetLocals(unittest.TestCase):
    """
    Test setting locals in one function from another function using several approaches.
    """

    def test_set_locals_using_save_locals(self):
        x = check_method(use_save_locals)
        self.assertEqual(x, 2)  # Expected to succeed


    def test_frame_simple_change(self):
        frame = sys._getframe()
        a = 20
        frame.f_locals['a'] = 50
        save_locals(frame)
        self.assertEqual(50, a)


    def test_frame_co_freevars(self):

        outer_var = 20

        def func():
            frame = sys._getframe()
            frame.f_locals['outer_var'] = 50
            save_locals(frame)
            self.assertEqual(50, outer_var)

        func()

    def test_frame_co_cellvars(self):

        def check_co_vars(a):
            frame = sys._getframe()
            def function2():
                print(a)

            assert 'a' in frame.f_code.co_cellvars
            frame = sys._getframe()
            frame.f_locals['a'] = 50
            save_locals(frame)
            self.assertEqual(50, a)

        check_co_vars(1)


    def test_frame_change_in_inner_frame(self):
        def change(f):
            self.assertTrue(f is not sys._getframe())
            f.f_locals['a']= 50
            save_locals(f)


        frame = sys._getframe()
        a = 20
        change(frame)
        self.assertEqual(50, a)


if __name__ == '__main__':
    suite = unittest.TestSuite()
#    suite.addTest(TestSetLocals('test_set_locals_using_dict'))
#    #suite.addTest(Test('testCase10a'))
#    unittest.TextTestRunner(verbosity=3).run(suite)

    suite = unittest.makeSuite(TestSetLocals)
    unittest.TextTestRunner(verbosity=3).run(suite)



class InvalidParameterValueError(Exception):
    def __init__(self, transformer, param, value, msg):
        exc_msg = f"{transformer}: Invalid '{param}' parameter value: '{value}'. {msg}"
        super().__init__(exc_msg)


class InvalidParameterError(Exception):
    def __init__(self, transformer, similar):
        super().__init__(
            f"{transformer}: Failed to import. Verify if correct name or configuration was provided.{similar}"
        )


class InvalidParameterFormatError(Exception):
    def __init__(self, transformer):
        super().__init__(
            f"{transformer}: Invalid parameter format. Pass parameters using MyTransformer:param_name=value syntax."
        )


class ImportTransformerError(ValueError):
    pass



# Original work Copyright Fabio Zadrozny (EPL 1.0)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains the main loops for our threads (reader_thread and writer_thread, which should
be initialized as the target of threading.Thread).

The reader_thread will read a message and ask a callable to deal with it and the
writer_thread just keeps on sending any message posted to the queue it receives.
"""

from functools import partial
import itertools
from robocorp_ls_core.robotframework_log import get_logger, get_log_level
import json
from typing import Optional, Dict


log = get_logger(__name__)

# Note: sentinel. Signals that the writer thread should stop processing.
STOP_WRITER_THREAD = "STOP_WRITER_THREAD"

# Note: sentinel. Sent by reader thread when stopped.
READER_THREAD_STOPPED = "READER_THREAD_STOPPED"


def read(stream, debug_prefix=b"read") -> Optional[Dict]:
    """
    Reads one message from the stream and returns the related dict (or None if EOF was reached).

    :param stream:
        The stream we should be reading from.

    :return dict|NoneType:
        The dict which represents a message or None if the stream was closed.
    """
    headers = {}
    while True:
        # Interpret the http protocol headers
        line = stream.readline()  # The trailing \r\n should be there.

        if get_log_level() > 1:
            log.debug(
                (
                    debug_prefix
                    + b": >>%s<<\n"
                    % (line.replace(b"\r", b"\\r").replace(b"\n", b"\\n"))
                ).decode("utf-8", "replace")
            )

        if not line:  # EOF
            return None
        line = line.strip().decode("ascii")
        if not line:  # Read just a new line without any contents
            break
        try:
            name, value = line.split(": ", 1)
        except ValueError:
            raise RuntimeError("Invalid header line: {}.".format(line))
        headers[name.strip()] = value.strip()

    if not headers:
        raise RuntimeError("Got message without headers.")

    content_length = int(headers["Content-Length"])

    # Get the actual json
    body = _read_len(stream, content_length)
    if get_log_level() > 1:
        log.debug((debug_prefix + b": %s" % (body,)).decode("utf-8", "replace"))

    try:
        return json.loads(body.decode("utf-8"))
    except:
        raise RuntimeError(f"Error reading: {body!r}")


def _read_len(stream, content_length) -> bytes:
    buf = b""
    if not content_length:
        return buf

    # Grab the body
    while True:
        data = stream.read(content_length - len(buf))
        if not buf and len(data) == content_length:
            # Common case
            return data
        buf += data
        if len(buf) == content_length:
            return buf
        if len(buf) > content_length:
            raise AssertionError(
                "Expected to read message up to len == %s (already read: %s). Found:\n%s"
                % (content_length, len(buf), buf.decode("utf-8", "replace"))
            )
        # len(buf) < content_length (just keep on going).


def reader_thread(
    stream,
    process_command,
    write_queue,
    debug_prefix=b"read",
    update_ids_from_dap=False,
):
    from robocorp_ls_core.debug_adapter_core.dap import dap_base_schema

    from robocorp_ls_core.debug_adapter_core.dap import (
        dap_schema,  # @UnusedImport -- register classes
    )
    from robocorp_ls_core.debug_adapter_core.dap.dap_schema import Response

    try:
        while True:
            data = read(stream, debug_prefix)
            if data is None:
                break
            try:

                # A response with success == False doesn't need to be translated
                # as the original response (to avoid the validation).
                if not data.get("success", True) and data.get("type") == "response":
                    protocol_message = dap_base_schema.from_dict(
                        data, update_ids_from_dap=update_ids_from_dap, cls=Response
                    )
                else:
                    protocol_message = dap_base_schema.from_dict(
                        data, update_ids_from_dap=update_ids_from_dap
                    )
                process_command(protocol_message)
            except Exception as e:
                log.exception("Error processing message.")
                seq = data.get("seq")
                if seq:
                    error_msg = {
                        "type": "response",
                        "request_seq": seq,
                        "success": False,
                        "command": data.get("command", "<unknown"),
                        "message": "Error processing message: %s" % (e,),
                    }
                    write_queue.put(error_msg)
    except ConnectionError:
        if get_log_level() > 2:
            log.exception("ConnectionError (ignored).")
    except:
        log.exception("Error reading message.")
    finally:
        process_command(READER_THREAD_STOPPED)


def writer_thread_no_auto_seq(
    stream, queue, debug_prefix="write", update_ids_to_dap=False
):
    """
    Same as writer_thread but does not set the message 'seq' automatically
    (meant to be used when responses, which need the seq id set need to be handled).
    """
    try:
        while True:
            to_write = queue.get()
            if to_write is STOP_WRITER_THREAD:
                log.debug("STOP_WRITER_THREAD")
                stream.close()
                break

            if isinstance(to_write, dict):
                assert "seq" in to_write
                try:
                    to_write = json.dumps(to_write)
                except:
                    log.exception("Error serializing %s to json.", to_write)
                    continue

            else:
                to_json = getattr(to_write, "to_json", None)
                if to_json is not None:
                    # Some protocol message
                    assert to_write.seq >= 0
                    try:
                        to_write = to_json(update_ids_to_dap=update_ids_to_dap)
                    except:
                        log.exception("Error serializing %s to json.", to_write)
                        continue

            if get_log_level() > 1:
                log.debug(debug_prefix + ": %s\n", to_write)

            if to_write.__class__ == bytes:
                as_bytes = to_write
            else:
                as_bytes = to_write.encode("utf-8")

            stream.write(
                ("Content-Length: %s\r\n\r\n" % (len(as_bytes))).encode("ascii")
            )
            stream.write(as_bytes)
            stream.flush()
    except:
        log.exception("Error writing message.")
    finally:
        log.debug("Exit reader thread.")


def writer_thread(stream, queue, debug_prefix="write", update_ids_to_dap=False):
    """
    Same as writer_thread_no_auto_seq but sets the message 'seq' automatically.
    """
    _next_seq = partial(next, itertools.count())

    try:
        while True:
            to_write = queue.get()
            if to_write is STOP_WRITER_THREAD:
                log.debug("STOP_WRITER_THREAD")
                stream.close()
                break

            if isinstance(to_write, dict):
                to_write["seq"] = _next_seq()
                try:
                    to_write = json.dumps(to_write)
                except:
                    log.exception("Error serializing %s to json.", to_write)
                    continue

            else:
                to_json = getattr(to_write, "to_json", None)
                if to_json is not None:
                    # Some protocol message
                    to_write.seq = _next_seq()
                    try:
                        to_write = to_json(update_ids_to_dap=update_ids_to_dap)
                    except:
                        log.exception("Error serializing %s to json.", to_write)
                        continue

            if get_log_level() > 1:
                log.debug(debug_prefix + ": %s\n", to_write)

            if to_write.__class__ == bytes:
                as_bytes = to_write
            else:
                as_bytes = to_write.encode("utf-8")

            stream.write(
                ("Content-Length: %s\r\n\r\n" % (len(as_bytes))).encode("ascii")
            )
            stream.write(as_bytes)
            stream.flush()
    except:
        log.exception("Error writing message.")
    finally:
        log.debug("Exit reader thread.")



from contextlib import contextmanager


def method1():

    _a = 0
    while _a < 2:  # break while
        _a += 1


def method2():
    try:
        raise AssertionError()
    except:  # break except
        pass


@contextmanager
def ctx():
    yield ''


def method3():
    with ctx() as a:  # break with
        return a


def method4():
    _a = 0
    for i in range(2):  # break for
        _a = i


def method5():
    try:  # break try 1
        _a = 10
    finally:
        _b = 10


def method6():
    try:
        _a = 10  # break try 2
    finally:
        _b = 10


def method7():
    try:
        _a = 10
    finally:
        _b = 10  # break finally 1


def method8():
    try:
        raise AssertionError()
    except:  # break except 2
        _b = 10
    finally:
        _c = 20


def method9():
    # As a note, Python 3.10 is eager to optimize this case and it duplicates the _c = 20
    # in a codepath where the exception is raised and another where it's not raised.
    # The frame eval mode must modify the bytecode so that both paths have the
    # programmatic breakpoint added!
    try:
        _a = 10
    except:
        _b = 10
    finally:_c = 20  # break finally 2


def method9a():
    # Same as method9, but with exception raised (but handled).
    try:
        raise AssertionError()
    except:
        _b = 10
    finally:_c = 20  # break finally 3


def method9b():
    # Same as method9, but with exception raised (unhandled).
    try:
        try:
            raise RuntimeError()
        except AssertionError:
            _b = 10
        finally:_c = 20  # break finally 4
    except:
        pass


def method10():
    _a = {
        0: 0,
        1: 1,  # break in dict
        2: 2,
    }


def method11():
    a = 11
    if a == 10:
        a = 20
    else: a = 30  # break else


if __name__ == '__main__':
    method1()
    method2()
    method3()
    method4()
    method5()
    method6()
    method7()
    method8()
    method9()
    method9a()
    method9b()
    method10()
    method11()
    print('TEST SUCEEDED')



"""
Main class of Robocop module. Gathers files to be scanned, checkers, parses CLI arguments and scans files.
"""
import inspect
import os
import sys
from collections import Counter

from robot.api import get_resource_model
from robot.errors import DataError

import robocop.exceptions
from robocop import checkers, reports
from robocop.rules import Message
from robocop.files import get_files
from robocop.config import Config
from robocop.utils import (
    DisablersFinder,
    FileType,
    FileTypeChecker,
    RecommendationFinder,
    is_suite_templated,
)


class Robocop:
    """
    Main class for running the checks.

    If you want to run checks with non default configuration create your own ``Config`` and pass it to ``Robocop``.
    Use ``Robocop.run()`` method to start analysis. If ``from_cli`` is set to ``False`` it will return
    list of found issues in JSON format.

    Example::

        import robocop
        from robocop.config import Config

        config = Config()
        config.include = {'1003'}
        config.paths = ['tests\\atest\\rules\\section-out-of-order']

        robocop_runner = robocop.Robocop(config=config)
        issues = robocop_runner.run()

    """

    def __init__(self, from_cli: bool = False, config: Config = None):
        self.files = {}
        self.checkers = []
        self.rules = {}
        self.reports = {}
        self.disabler = None
        self.root = os.getcwd()
        self.config = Config(from_cli=from_cli) if config is None else config
        self.from_cli = from_cli
        if not from_cli:
            self.config.reports.add("json_report")
        self.out = self.set_output()

    def set_output(self):
        """Set output for printing to file if configured. Else use standard output"""
        return self.config.output or None

    def write_line(self, line):
        """Print line using file=self.out parameter (set in `set_output` method)"""
        print(line, file=self.out)

    def reload_config(self):
        """Reload checkers and reports based on current config"""
        self.load_checkers()
        self.config.validate_rule_names(self.rules)
        self.list_checkers()
        self.load_reports()
        self.configure_checkers_or_reports()

    def run(self):
        """Entry point for running scans"""
        self.reload_config()
        self.recognize_file_types()
        self.run_checks()
        self.make_reports()
        if self.config.output and not self.out.closed:
            self.out.close()
        if self.from_cli:
            sys.exit(self.reports["return_status"].return_status)
        else:
            return self.reports["json_report"].issues

    def recognize_file_types(self):
        """
        Pre-parse files to recognize their types. If the filename is `__init__.*`, the type is `INIT`.
        Files with .resource extension are `RESOURCE` type.
        If the file is imported somewhere then file type is `RESOURCE`. Otherwise file type is `GENERAL`.
        These types are important since they are used to define parsing class for robot API.
        """
        file_type_checker = FileTypeChecker(self.config.exec_dir)
        for file in get_files(self.config):
            if "__init__" in file.name:
                file_type = FileType.INIT
            elif file.suffix.lower() == ".resource":
                file_type = FileType.RESOURCE
            else:
                file_type = FileType.GENERAL
            file_type_checker.source = file
            try:
                model = file_type.get_parser()(str(file))
                file_type_checker.visit(model)
                self.files[file] = (file_type, model)
            except DataError:
                print(f"Failed to decode {file}. Default supported encoding by Robot Framework is UTF-8. Skipping file")

        for resource in file_type_checker.resource_files:
            if resource in self.files and self.files[resource][0].value != FileType.RESOURCE:
                self.files[resource] = (
                    FileType.RESOURCE,
                    get_resource_model(str(resource)),
                )

    def run_checks(self):
        for file in self.files:
            if self.config.verbose:
                print(f"Scanning file: {file}")
            model = self.files[file][1]
            found_issues = self.run_check(model, str(file))
            found_issues.sort()
            for issue in found_issues:
                self.report(issue)
        if "file_stats" in self.reports:
            self.reports["file_stats"].files_count = len(self.files)

    def run_check(self, ast_model, filename, source=None):
        found_issues = []
        self.register_disablers(filename, source)
        if self.disabler.file_disabled:
            return []
        templated = is_suite_templated(ast_model)
        for checker in self.checkers:
            if checker.disabled:
                continue
            found_issues += [
                issue
                for issue in checker.scan_file(ast_model, filename, source, templated)
                if not self.disabler.is_rule_disabled(issue)
            ]
        return found_issues

    def register_disablers(self, filename, source):
        """Parse content of file to find any disabler statements like # robocop: disable=rulename"""
        self.disabler = DisablersFinder(filename=filename, source=source)

    def report(self, rule_msg: Message):
        for report in self.reports.values():
            report.add_message(rule_msg)
        try:
            source_rel = os.path.relpath(os.path.expanduser(rule_msg.source), self.root)
        except ValueError:
            source_rel = rule_msg.source
        self.log_message(
            source=rule_msg.source,
            source_rel=source_rel,
            line=rule_msg.line,
            col=rule_msg.col,
            end_line=rule_msg.end_line,
            end_col=rule_msg.end_col,
            severity=rule_msg.severity.value,
            rule_id=rule_msg.rule_id,
            desc=rule_msg.desc,
            name=rule_msg.name,
        )

    def log_message(self, **kwargs):
        self.write_line(self.config.format.format(**kwargs))

    def load_checkers(self):
        self.checkers = []
        self.rules = {}
        checkers.init(self)

    def list_checkers(self):
        if not (self.config.list or self.config.list_configurables):
            return
        if self.config.list_configurables:
            print(
                "All rules have configurable parameter 'severity'. Allowed values are:"
                "\n    E / error\n    W / warning\n    I / info"
            )
        pattern = self.config.list if self.config.list else self.config.list_configurables
        rule_by_id = {rule.rule_id: rule for rule in self.rules.values() if rule.matches_pattern(pattern)}
        rule_by_id = sorted(rule_by_id.values(), key=lambda x: x.rule_id)
        severity_counter = Counter({"E": 0, "W": 0, "I": 0})
        for rule in rule_by_id:
            if self.config.list:
                print(rule)
                severity_counter[rule.severity.value] += 1
            else:
                params = rule.available_configurables(include_severity=False)
                if params:
                    print(f"{rule}\n" f"    {params}")
                    severity_counter[rule.severity.value] += 1
        print(
            f"\nAltogether {sum(severity_counter.values())} rule(s) with following severity:\n"
            f"    {severity_counter['E']} error rule(s),\n"
            f"    {severity_counter['W']} warning rule(s),\n"
            f"    {severity_counter['I']} info rule(s).\n"
        )
        print("Visit https://robocop.readthedocs.io/en/stable/rules.html page for detailed documentation.")
        sys.exit()

    def load_reports(self):
        self.reports = {}
        classes = inspect.getmembers(reports, inspect.isclass)
        available_reports = "Available reports:\n"
        for report_class in classes:
            if not issubclass(report_class[1], reports.Report):
                continue
            report = report_class[1]()
            if not hasattr(report, "name"):
                continue
            if "all" in self.config.reports or report.name in self.config.reports:
                self.reports[report.name] = report
            available_reports += f"{report.name:20} - {report.description}\n"
        if self.config.list_reports:
            available_reports += "all" + " " * 18 + "- Turns on all available reports"
            print(available_reports)
            sys.exit()

    def register_checker(self, checker):
        if not self.any_rule_enabled(checker):
            checker.disabled = True
        for rule_name, rule in checker.rules.items():
            self.rules[rule_name] = rule
            self.rules[rule.rule_id] = rule
        self.checkers.append(checker)

    def make_reports(self):
        for report in self.reports.values():
            output = report.get_report()
            if output is not None:
                self.write_line(output)

    def any_rule_enabled(self, checker) -> bool:
        for name, rule in checker.rules.items():
            rule.enabled = self.config.is_rule_enabled(rule)
            checker.rules[name] = rule
        return any(msg.enabled for msg in checker.rules.values())

    def configure_checkers_or_reports(self):
        for config in self.config.configure:
            if config.count(":") < 2:
                raise robocop.exceptions.ConfigGeneralError(
                    f"Provided invalid config: '{config}' (general pattern: <rule>:<param>:<value>)"
                )
            rule_or_report, param, value = config.split(":", maxsplit=2)
            if rule_or_report in self.rules:
                rule = self.rules[rule_or_report]
                rule.configure(param, value)
            elif rule_or_report in self.reports:
                self.reports[rule_or_report].configure(param, value)
            else:
                similar = RecommendationFinder().find_similar(rule_or_report, self.rules)
                raise robocop.exceptions.ConfigGeneralError(
                    f"Provided rule or report '{rule_or_report}' does not exist. {similar}"
                )


def run_robocop():
    try:
        linter = Robocop(from_cli=True)
        linter.run()
    except robocop.exceptions.RobocopFatalError as err:
        print(f"Error: {err}")
        sys.exit(1)
    except Exception as err:
        message = (
            "\nFatal exception occurred. You can create an issue at "
            "https://github.com/MarketSquare/robotframework-robocop/issues/new/choose . Thanks!"
        )
        err.args = (err.args[0] + message,) + err.args[1:]
        raise err



class _Callback(object):
    def __init__(self):
        self._callbacks = []

    def register(self, callback):
        self._callbacks.append(callback)

    def __call__(self, *args, **kwargs):
        for c in self._callbacks:
            c(*args, **kwargs)


class DebugListener(object):
    ROBOT_LISTENER_API_VERSION = 3

    on_start_suite = _Callback()
    on_end_suite = _Callback()
    on_start_test = _Callback()
    on_end_test = _Callback()
    on_log_message = _Callback()
    on_message = _Callback()

    def start_suite(self, data, result):
        self.on_start_suite(data, result)

    def end_suite(self, data, result):
        self.on_end_suite(data, result)

    def start_test(self, data, result):
        self.on_start_test(data, result)

    def end_test(self, data, result):
        self.on_end_test(data, result)

    def log_message(self, message):
        self.on_log_message(message)

    def message(self, message):
        self.on_message(message)

    # def start_keyword(self, data, result):
    #     # This would be nice, but it's not currently supported.
    #
    # def end_keyword(self, data, result):
    #     # This would be nice, but it's not currently supported.


class DebugListenerV2(object):
    ROBOT_LISTENER_API_VERSION = 2

    on_start_keyword = _Callback()
    on_end_keyword = _Callback()

    def start_suite(self, name, attributes):
        pass

    def end_suite(self, name, attributes):
        pass

    def start_test(self, name, attributes):
        pass

    def end_test(self, name, attributes):
        pass

    def start_keyword(self, name, attributes):
        self.on_start_keyword(name, attributes)

    def end_keyword(self, name, attributes):
        self.on_end_keyword(name, attributes)



# Hack so that we don't break the runtime on versions prior to Python 3.8.
import sys
from typing import TypeVar, List, Union, Any, Optional, Iterable
from robocorp_ls_core.debug_adapter_core.dap.dap_schema import StackFrame
from robocorp_ls_core.protocols import IFuture

if sys.version_info[:2] < (3, 8):

    class Protocol(object):
        pass


else:
    from typing import Protocol

T = TypeVar("T")
Y = TypeVar("Y", covariant=True)


class IEvaluationInfo(Protocol):
    future: IFuture[Any]


class IRobotBreakpoint(Protocol):
    # 1-based line for the breakpoint.
    lineno: int
    condition: Optional[str]
    hit_condition: Optional[int]
    log_message: Optional[str]

    hits: int


class IBusyWait(Protocol):
    before_wait: List[Any]
    waited: int
    proceeded: int

    def pre_wait(self) -> None:
        pass

    def wait(self) -> None:
        pass

    def proceed(self) -> None:
        pass


class IRobotDebugger(Protocol):
    busy_wait: IBusyWait

    def reset(self):
        pass

    def evaluate(
        self, frame_id: int, expression: str, context: str = "watch"
    ) -> IEvaluationInfo:
        """
        Asks something to be evaluated.

        This is an asynchronous operation and returns an _EvaluationInfo (to get
        the result, access _EvaluationInfo.future.result())
        """

    def step_continue(self) -> None:
        pass

    def step_in(self) -> None:
        pass

    def step_next(self) -> None:
        pass

    def step_out(self) -> None:
        pass

    def set_breakpoints(
        self,
        filename: str,
        breakpoints: Union[IRobotBreakpoint, Iterable[IRobotBreakpoint]],
    ) -> None:
        """
        :param str filename:
        :param list(RobotBreakpoint) breakpoints:
        """

    def get_frames(self, thread_id) -> Optional[List[StackFrame]]:
        pass

    def iter_frame_ids(self, thread_id) -> Iterable[int]:
        pass

    def get_current_thread_id(self, thread=None) -> int:
        pass

    def write_message(self, msg):
        pass


class INextId(Protocol):
    def __call__(self) -> T:
        pass



from typing import Optional, Tuple


def extract_source_and_line_from_message(msg) -> Optional[Tuple[str, int]]:
    """
    :param msg:
        Something as:
        Error in file 'C:/Users/.../case_import_failure.robot' on line 2
    """
    if msg:
        first_line = msg.split("\n")[0]
        import re

        m = re.search(r"file\s'(.*)'\son\sline\s(\d+)", first_line)
        if m:
            return m.group(1), int(m.group(2))
    return None



# Original work Copyright Fabio Zadrozny (EPL 1.0)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Run this module to regenerate the `dap_schema.py` file.

Note that it'll generate it based on the current debugProtocol.json. Erase it and rerun
to download the latest version.

Based on: https://github.com/fabioz/PyDev.Debugger/blob/master/_pydevd_bundle/_debug_adapter/__main__pydevd_gen_debug_adapter_protocol.py
"""


def is_variable_to_translate(cls_name, var_name):
    if var_name in ("variablesReference", "frameId", "threadId"):
        return True

    if cls_name == "StackFrame" and var_name == "id":
        # It's frameId everywhere except on StackFrame.
        return True

    if cls_name == "Thread" and var_name == "id":
        # It's threadId everywhere except on Thread.
        return True

    return False


def _get_noqa_for_var(prop_name):
    return (
        "  # noqa (assign to builtin)"
        if prop_name in ("type", "format", "id", "hex", "breakpoint", "filter")
        else ""
    )


class _OrderedSet(object):
    # Not a good ordered set (just something to be small without adding any deps)

    def __init__(self, initial_contents=None):
        self._contents = []
        self._contents_as_set = set()
        if initial_contents is not None:
            for x in initial_contents:
                self.add(x)

    def add(self, x):
        if x not in self._contents_as_set:
            self._contents_as_set.add(x)
            self._contents.append(x)

    def discard(self, x):
        if x in self._contents_as_set:
            self._contents_as_set.remove(x)
            self._contents.remove(x)

    def copy(self):
        return _OrderedSet(self._contents)

    def update(self, contents):
        for x in contents:
            self.add(x)

    def __iter__(self):
        return iter(self._contents)

    def __contains__(self, item):
        return item in self._contents_as_set

    def __len__(self):
        return len(self._contents)

    def set_repr(self):
        if len(self) == 0:
            return "set()"

        lst = [repr(x) for x in self]
        return "set([" + ", ".join(lst) + "])"


class Ref(object):
    def __init__(self, ref, ref_data):
        self.ref = ref
        self.ref_data = ref_data

    def __str__(self):
        return self.ref


def load_schema_data():
    import os.path
    import json

    json_file = os.path.join(os.path.dirname(__file__), "debugProtocol.json")
    if not os.path.exists(json_file):
        import requests

        req = requests.get(
            "https://raw.githubusercontent.com/microsoft/debug-adapter-protocol/gh-pages/debugAdapterProtocol.json"
        )
        assert req.status_code == 200
        with open(json_file, "wb") as stream:
            stream.write(req.content)

    with open(json_file, "rb") as json_contents:
        json_schema_data = json.loads(json_contents.read())
    return json_schema_data


def load_custom_schema_data():
    import os.path
    import json

    json_file = os.path.join(os.path.dirname(__file__), "debugProtocolCustom.json")

    with open(json_file, "rb") as json_contents:
        json_schema_data = json.loads(json_contents.read())
    return json_schema_data


def create_classes_to_generate_structure(json_schema_data):
    definitions = json_schema_data["definitions"]

    class_to_generatees = {}

    for name, definition in definitions.items():
        all_of = definition.get("allOf")
        description = definition.get("description")
        is_enum = definition.get("type") == "string" and "enum" in definition
        enum_values = None
        if is_enum:
            enum_values = definition["enum"]
        properties = {}
        properties.update(definition.get("properties", {}))
        required = _OrderedSet(definition.get("required", _OrderedSet()))
        base_definitions = []

        if all_of is not None:
            for definition in all_of:
                ref = definition.get("$ref")
                if ref is not None:
                    assert ref.startswith("#/definitions/")
                    ref = ref[len("#/definitions/") :]
                    base_definitions.append(ref)
                else:
                    if not description:
                        description = definition.get("description")
                    properties.update(definition.get("properties", {}))
                    required.update(
                        _OrderedSet(definition.get("required", _OrderedSet()))
                    )

        if isinstance(description, (list, tuple)):
            description = "\n".join(description)

        if (
            name == "ModulesRequest"
        ):  # Hack to accept modules request without arguments (ptvsd: 2050).
            required.discard("arguments")
        class_to_generatees[name] = dict(
            name=name,
            properties=properties,
            base_definitions=base_definitions,
            description=description,
            required=required,
            is_enum=is_enum,
            enum_values=enum_values,
        )
    return class_to_generatees


def collect_bases(curr_class, classes_to_generate, memo=None):
    ret = []
    if memo is None:
        memo = {}

    base_definitions = curr_class["base_definitions"]
    for base_definition in base_definitions:
        if base_definition not in memo:
            ret.append(base_definition)
            ret.extend(
                collect_bases(
                    classes_to_generate[base_definition], classes_to_generate, memo
                )
            )

    return ret


def fill_properties_and_required_from_base(classes_to_generate):
    # Now, resolve properties based on refs
    for class_to_generate in classes_to_generate.values():
        dct = {}
        s = _OrderedSet()

        for base_definition in reversed(
            collect_bases(class_to_generate, classes_to_generate)
        ):
            # Note: go from base to current so that the initial order of the properties has that
            # same order.
            dct.update(classes_to_generate[base_definition].get("properties", {}))
            s.update(
                classes_to_generate[base_definition].get("required", _OrderedSet())
            )

        dct.update(class_to_generate["properties"])
        class_to_generate["properties"] = dct

        s.update(class_to_generate["required"])
        class_to_generate["required"] = s

    return class_to_generate


def update_class_to_generate_description(class_to_generate):
    import textwrap

    description = class_to_generate["description"]
    lines = []
    for line in description.splitlines():
        wrapped = textwrap.wrap(line.strip(), 100)
        lines.extend(wrapped)
        lines.append("")

    while lines and lines[-1] == "":
        lines = lines[:-1]

    class_to_generate["description"] = "    " + ("\n    ".join(lines))


def update_class_to_generate_type(classes_to_generate, class_to_generate):
    properties = class_to_generate.get("properties")
    for _prop_name, prop_val in properties.items():
        prop_type = prop_val.get("type", "")
        if not prop_type:
            prop_type = prop_val.pop("$ref", "")
            if prop_type:
                assert prop_type.startswith("#/definitions/")
                prop_type = prop_type[len("#/definitions/") :]
                prop_val["type"] = Ref(prop_type, classes_to_generate[prop_type])


def update_class_to_generate_register_dec(classes_to_generate, class_to_generate):
    # Default
    class_to_generate["register_request"] = ""
    class_to_generate["register_dec"] = "@register"

    properties = class_to_generate.get("properties")
    enum_type = properties.get("type", {}).get("enum")
    command = None
    event = None
    if (
        enum_type
        and len(enum_type) == 1
        and next(iter(enum_type)) in ("request", "response", "event")
    ):
        msg_type = next(iter(enum_type))
        if msg_type == "response":
            # The actual command is typed in the request
            response_name = class_to_generate["name"]
            request_name = response_name[: -len("Response")] + "Request"
            if request_name in classes_to_generate:
                command = classes_to_generate[request_name]["properties"].get("command")
            else:
                if response_name == "ErrorResponse":
                    command = {"enum": ["error"]}
                else:
                    raise AssertionError("Unhandled: %s" % (response_name,))

        elif msg_type == "request":
            command = properties.get("command")

        elif msg_type == "event":
            command = properties.get("event")

        else:
            raise AssertionError("Unexpected condition.")

        if command:
            enum = command.get("enum")
            if enum and len(enum) == 1:
                class_to_generate["register_request"] = "@register_%s(%r)\n" % (
                    msg_type,
                    enum[0],
                )


def extract_prop_name_and_prop(class_to_generate):
    properties = class_to_generate.get("properties")
    required = _OrderedSet(class_to_generate.get("required", _OrderedSet()))

    # Sort so that required come first
    prop_name_and_prop = list(properties.items())

    def compute_sort_key(x):
        key = x[0]
        if key in required:
            if key == "seq":
                return 0.5  # seq when required is after the other required keys (to have a default of -1).
            return 0
        return 1

    prop_name_and_prop.sort(key=compute_sort_key)

    return prop_name_and_prop


def update_class_to_generate_to_json(class_to_generate):
    required = _OrderedSet(class_to_generate.get("required", _OrderedSet()))
    prop_name_and_prop = extract_prop_name_and_prop(class_to_generate)

    to_dict_body = [
        "def to_dict(self, update_ids_to_dap=False):  # noqa (update_ids_to_dap may be unused)"
    ]

    translate_prop_names = []
    for prop_name, prop in prop_name_and_prop:
        if is_variable_to_translate(class_to_generate["name"], prop_name):
            translate_prop_names.append(prop_name)

    for prop_name, prop in prop_name_and_prop:
        namespace = dict(prop_name=prop_name, noqa=_get_noqa_for_var(prop_name))
        to_dict_body.append(
            "    %(prop_name)s = self.%(prop_name)s%(noqa)s" % namespace
        )

        if prop.get("type") == "array":
            to_dict_body.append(
                '    if %(prop_name)s and hasattr(%(prop_name)s[0], "to_dict"):'
                % namespace
            )
            to_dict_body.append(
                "        %(prop_name)s = [x.to_dict() for x in %(prop_name)s]"
                % namespace
            )

    if translate_prop_names:
        to_dict_body.append("    if update_ids_to_dap:")
        for prop_name in translate_prop_names:
            namespace = dict(prop_name=prop_name, noqa=_get_noqa_for_var(prop_name))
            to_dict_body.append("        if %(prop_name)s is not None:" % namespace)
            to_dict_body.append(
                "            %(prop_name)s = self._translate_id_to_dap(%(prop_name)s)%(noqa)s"
                % namespace
            )

    if not translate_prop_names:
        update_dict_ids_from_dap_body = []
    else:
        update_dict_ids_from_dap_body = [
            "",
            "",
            "@classmethod",
            "def update_dict_ids_from_dap(cls, dct):",
        ]
        for prop_name in translate_prop_names:
            namespace = dict(prop_name=prop_name)
            update_dict_ids_from_dap_body.append(
                "    if %(prop_name)r in dct:" % namespace
            )
            update_dict_ids_from_dap_body.append(
                "        dct[%(prop_name)r] = cls._translate_id_from_dap(dct[%(prop_name)r])"
                % namespace
            )
        update_dict_ids_from_dap_body.append("    return dct")

    class_to_generate["update_dict_ids_from_dap"] = _indent_lines(
        "\n".join(update_dict_ids_from_dap_body)
    )

    to_dict_body.append("    dct = {")
    first_not_required = False

    for prop_name, prop in prop_name_and_prop:
        use_to_dict = prop["type"].__class__ == Ref and not prop["type"].ref_data.get(
            "is_enum", False
        )
        is_array = prop["type"] == "array"
        ref_array_cls_name = ""
        if is_array:
            ref = prop["items"].get("$ref")
            if ref is not None:
                ref_array_cls_name = ref.split("/")[-1]

        namespace = dict(prop_name=prop_name, ref_array_cls_name=ref_array_cls_name)
        if prop_name in required:
            if use_to_dict:
                to_dict_body.append(
                    "        %(prop_name)r: %(prop_name)s.to_dict(update_ids_to_dap=update_ids_to_dap),"
                    % namespace
                )
            else:
                if ref_array_cls_name:
                    to_dict_body.append(
                        "        %(prop_name)r: [%(ref_array_cls_name)s.update_dict_ids_to_dap(o) for o in %(prop_name)s] if (update_ids_to_dap and %(prop_name)s) else %(prop_name)s,"
                        % namespace
                    )
                else:
                    to_dict_body.append(
                        "        %(prop_name)r: %(prop_name)s," % namespace
                    )
        else:
            if not first_not_required:
                first_not_required = True
                to_dict_body.append("    }")

            to_dict_body.append("    if %(prop_name)s is not None:" % namespace)
            if use_to_dict:
                to_dict_body.append(
                    "        dct[%(prop_name)r] = %(prop_name)s.to_dict(update_ids_to_dap=update_ids_to_dap)"
                    % namespace
                )
            else:
                if ref_array_cls_name:
                    to_dict_body.append(
                        "        dct[%(prop_name)r] = [%(ref_array_cls_name)s.update_dict_ids_to_dap(o) for o in %(prop_name)s] if (update_ids_to_dap and %(prop_name)s) else %(prop_name)s"
                        % namespace
                    )
                else:
                    to_dict_body.append(
                        "        dct[%(prop_name)r] = %(prop_name)s" % namespace
                    )

    if not first_not_required:
        first_not_required = True
        to_dict_body.append("    }")

    to_dict_body.append("    dct.update(self.kwargs)")
    to_dict_body.append("    return dct")

    class_to_generate["to_dict"] = _indent_lines("\n".join(to_dict_body))

    if not translate_prop_names:
        update_dict_ids_to_dap_body = []
    else:
        update_dict_ids_to_dap_body = [
            "",
            "",
            "@classmethod",
            "def update_dict_ids_to_dap(cls, dct):",
        ]
        for prop_name in translate_prop_names:
            namespace = dict(prop_name=prop_name)
            update_dict_ids_to_dap_body.append(
                "    if %(prop_name)r in dct:" % namespace
            )
            update_dict_ids_to_dap_body.append(
                "        dct[%(prop_name)r] = cls._translate_id_to_dap(dct[%(prop_name)r])"
                % namespace
            )
        update_dict_ids_to_dap_body.append("    return dct")

    class_to_generate["update_dict_ids_to_dap"] = _indent_lines(
        "\n".join(update_dict_ids_to_dap_body)
    )


def update_class_to_generate_init(class_to_generate):
    args = []
    init_body = []
    docstring = []

    required = _OrderedSet(class_to_generate.get("required", _OrderedSet()))
    prop_name_and_prop = extract_prop_name_and_prop(class_to_generate)

    translate_prop_names = []
    for prop_name, prop in prop_name_and_prop:
        if is_variable_to_translate(class_to_generate["name"], prop_name):
            translate_prop_names.append(prop_name)

        enum = prop.get("enum")
        if enum and len(enum) == 1:
            init_body.append(
                "    self.%(prop_name)s = %(enum)r"
                % dict(prop_name=prop_name, enum=next(iter(enum)))
            )
        else:
            if prop_name in required:
                if prop_name == "seq":
                    args.append(prop_name + "=-1")
                else:
                    args.append(prop_name)
            else:
                args.append(prop_name + "=None")

            if prop["type"].__class__ == Ref:
                ref = prop["type"]
                ref_data = ref.ref_data
                if ref_data.get("is_enum", False):
                    init_body.append(
                        "    assert %s in %s.VALID_VALUES" % (prop_name, str(ref))
                    )
                    init_body.append(
                        "    self.%(prop_name)s = %(prop_name)s"
                        % dict(prop_name=prop_name)
                    )
                else:
                    namespace = dict(prop_name=prop_name, ref_name=str(ref))
                    init_body.append("    if %(prop_name)s is None:" % namespace)
                    init_body.append(
                        "        self.%(prop_name)s = %(ref_name)s()" % namespace
                    )
                    init_body.append("    else:")
                    init_body.append(
                        "        self.%(prop_name)s = %(ref_name)s(update_ids_from_dap=update_ids_from_dap, **%(prop_name)s) if %(prop_name)s.__class__ !=  %(ref_name)s else %(prop_name)s"
                        % namespace
                    )

            else:
                init_body.append(
                    "    self.%(prop_name)s = %(prop_name)s" % dict(prop_name=prop_name)
                )

                if prop["type"] == "array":
                    ref = prop["items"].get("$ref")
                    if ref is not None:
                        ref_array_cls_name = ref.split("/")[-1]
                        init_body.append(
                            "    if update_ids_from_dap and self.%(prop_name)s:"
                            % dict(prop_name=prop_name)
                        )
                        init_body.append(
                            "        for o in self.%(prop_name)s:"
                            % dict(prop_name=prop_name)
                        )
                        init_body.append(
                            "            %(ref_array_cls_name)s.update_dict_ids_from_dap(o)"
                            % dict(ref_array_cls_name=ref_array_cls_name)
                        )

        prop_type = prop["type"]
        prop_description = prop.get("description", "")

        if isinstance(prop_description, (list, tuple)):
            prop_description = "\n    ".join(prop_description)

        docstring.append(
            ":param %(prop_type)s %(prop_name)s: %(prop_description)s"
            % dict(
                prop_type=prop_type,
                prop_name=prop_name,
                prop_description=prop_description,
            )
        )

    if translate_prop_names:
        init_body.append("    if update_ids_from_dap:")
        for prop_name in translate_prop_names:
            init_body.append(
                "        self.%(prop_name)s = self._translate_id_from_dap(self.%(prop_name)s)"
                % dict(prop_name=prop_name)
            )

    docstring = _indent_lines("\n".join(docstring))
    init_body = "\n".join(init_body)

    # Actually bundle the whole __init__ from the parts.
    args = ", ".join(args)
    if args:
        args = ", " + args

    # Note: added kwargs because some messages are expected to be extended by the user (so, we'll actually
    # make all extendable so that we don't have to worry about which ones -- we loose a little on typing,
    # but may be better than doing a whitelist based on something only pointed out in the documentation).
    class_to_generate[
        "init"
    ] = '''def __init__(self%(args)s, update_ids_from_dap=False, **kwargs):  # noqa (update_ids_from_dap may be unused)
    """
%(docstring)s
    """
%(init_body)s
    self.kwargs = kwargs
''' % dict(
        args=args, init_body=init_body, docstring=docstring
    )

    class_to_generate["init"] = _indent_lines(class_to_generate["init"])


def update_class_to_generate_props(class_to_generate):
    import json

    def default(o):
        if isinstance(o, Ref):
            return o.ref
        raise AssertionError("Unhandled: %s" % (o,))

    properties = class_to_generate["properties"]
    class_to_generate["props"] = (
        "    __props__ = %s"
        % _indent_lines(json.dumps(properties, indent=4, default=default)).strip()
    )


def update_class_to_generate_refs(class_to_generate):
    properties = class_to_generate["properties"]
    class_to_generate["refs"] = (
        "    __refs__ = %s"
        % _OrderedSet(
            key for (key, val) in properties.items() if val["type"].__class__ == Ref
        ).set_repr()
    )


def update_class_to_generate_enums(class_to_generate):
    class_to_generate["enums"] = ""
    if class_to_generate.get("is_enum", False):
        enums = ""
        for enum in class_to_generate["enum_values"]:
            enums += "    %s = %r\n" % (enum.upper(), enum)
        enums += "\n"
        enums += (
            "    VALID_VALUES = %s\n\n"
            % _OrderedSet(class_to_generate["enum_values"]).set_repr()
        )
        class_to_generate["enums"] = enums


def update_class_to_generate_objects(classes_to_generate, class_to_generate):
    properties = class_to_generate["properties"]
    for key, val in properties.items():
        if val["type"] == "object":
            create_new = val.copy()
            create_new.update(
                {
                    "name": "%s%s" % (class_to_generate["name"], key.title()),
                    "description": '    "%s" of %s' % (key, class_to_generate["name"]),
                }
            )
            if "properties" not in create_new:
                create_new["properties"] = {}

            assert create_new["name"] not in classes_to_generate
            classes_to_generate[create_new["name"]] = create_new

            update_class_to_generate_type(classes_to_generate, create_new)
            update_class_to_generate_props(create_new)

            # Update nested object types
            update_class_to_generate_objects(classes_to_generate, create_new)

            val["type"] = Ref(
                create_new["name"], classes_to_generate[create_new["name"]]
            )
            val.pop("properties", None)


def gen_debugger_protocol():
    import os.path
    import sys

    if sys.version_info[:2] < (3, 6):
        raise AssertionError(
            "Must be run with Python 3.6 onwards (to keep dict order)."
        )

    classes_to_generate = create_classes_to_generate_structure(load_schema_data())
    classes_to_generate.update(
        create_classes_to_generate_structure(load_custom_schema_data())
    )

    class_to_generate = fill_properties_and_required_from_base(classes_to_generate)

    for class_to_generate in list(classes_to_generate.values()):
        update_class_to_generate_description(class_to_generate)
        update_class_to_generate_type(classes_to_generate, class_to_generate)
        update_class_to_generate_props(class_to_generate)
        update_class_to_generate_objects(classes_to_generate, class_to_generate)

    for class_to_generate in classes_to_generate.values():
        update_class_to_generate_refs(class_to_generate)
        update_class_to_generate_init(class_to_generate)
        update_class_to_generate_enums(class_to_generate)
        update_class_to_generate_to_json(class_to_generate)
        update_class_to_generate_register_dec(classes_to_generate, class_to_generate)

    class_template = '''
%(register_request)s%(register_dec)s
class %(name)s(BaseSchema):
    """
%(description)s

    Note: automatically generated code. Do not edit manually.
    """

%(enums)s%(props)s
%(refs)s

    __slots__ = list(__props__.keys()) + ['kwargs']

%(init)s%(update_dict_ids_from_dap)s

%(to_dict)s%(update_dict_ids_to_dap)s
'''

    contents = []
    contents.append("# coding: utf-8")
    contents.append("# fmt: off")
    contents.append("# Automatically generated code.")
    contents.append("# Do not edit manually.")
    contents.append("# Generated by running: %s" % os.path.basename(__file__))
    contents.append(
        "from .dap_base_schema import BaseSchema, register, register_request, register_response, register_event"
    )
    contents.append("")
    for class_to_generate in classes_to_generate.values():
        contents.append(class_template % class_to_generate)

    contents.append("# fmt: on")
    contents.append("")
    parent_dir = os.path.dirname(__file__)
    schema = os.path.join(parent_dir, "dap_schema.py")
    with open(schema, "w", encoding="utf-8") as stream:
        stream.write("\n".join(contents))


def _indent_lines(lines, indent="    "):
    out_lines = []
    for line in lines.splitlines(keepends=True):
        out_lines.append(indent + line)

    return "".join(out_lines)


if __name__ == "__main__":

    gen_debugger_protocol()



import subprocess


def _fix_subprocess():
    import sys
    from robocorp_ls_core.constants import IS_PY37_ONWARDS

    # Workaround for https://bugs.python.org/issue37380 in older versions of Python.
    # i.e.: OSError: [WinError 6] The handle is invalid
    if IS_PY37_ONWARDS:
        return  # This is fixed in 3.7 onwards.

    _cleanup = (
        subprocess._cleanup
    )  # Just check that there is in fact a _cleanup attribute there.

    def _new_cleanup():
        for inst in subprocess._active[:]:
            try:
                res = inst._internal_poll(_deadstate=sys.maxsize)
            except OSError:
                res = 1  # This is the fix.
            if res is not None:
                try:
                    subprocess._active.remove(inst)
                except ValueError:
                    # This can happen if two threads create a new Popen instance.
                    # It's harmless that it was already removed, so ignore.
                    pass

    subprocess._cleanup = _new_cleanup


_fix_subprocess()



def Call():
    b = True
    while b:  # expected
        # requested
        pass  # Note: until 3.10 a pass didn't generate a line event, but starting at 3.10, it does...
        break


if __name__ == '__main__':
    Call()
    print('TEST SUCEEDED!')



# coding: utf-8
import collections
import sys
import re
import pytest
from _pydevd_bundle.pydevd_safe_repr import SafeRepr
import json
from _pydevd_bundle.pydevd_constants import IS_PY36_OR_GREATER

try:
    import numpy as np
except ImportError:
    np = None

PY_VER = sys.version_info[0]
assert PY_VER <= 3  # Fix the code when Python 4 comes around.
PY3K = PY_VER == 3


class SafeReprTestBase(object):

    saferepr = SafeRepr()

    def assert_saferepr(self, value, expected):
        safe = self.saferepr(value)

        if len(safe) != len(expected):
            raise AssertionError('Expected:\n%s\nFound:\n%s\n Expected len: %s Found len: %s' % (
                expected, safe, len(expected), len(safe),))
        assert safe == expected
        return safe

    def assert_unchanged(self, value, expected):
        actual = repr(value)

        safe = self.assert_saferepr(value, expected)
        assert safe == actual

    def assert_shortened(self, value, expected):
        actual = repr(value)

        safe = self.assert_saferepr(value, expected)
        assert safe != actual

    def assert_saferepr_regex(self, s, r):
        safe = self.saferepr(s)

        assert re.search(r, safe) is not None
        return safe

    def assert_unchanged_regex(self, value, expected):
        actual = repr(value)

        safe = self.assert_saferepr_regex(value, expected)
        assert safe == actual

    def assert_shortened_regex(self, value, expected):
        actual = repr(value)

        safe = self.assert_saferepr_regex(value, expected)
        assert safe != actual


class TestSafeRepr(SafeReprTestBase):

    def test_collection_types(self):
        colltypes = [t for t, _, _, _ in SafeRepr.collection_types]

        assert colltypes == [
            tuple,
            list,
            frozenset,
            set,
            collections.deque,
        ]

    def test_largest_repr(self):
        # Find the largest possible repr and ensure it is below our arbitrary
        # limit (8KB).
        coll = '-' * (SafeRepr.maxstring_outer * 2)
        for limit in reversed(SafeRepr.maxcollection[1:]):
            coll = [coll] * (limit * 2)
        dcoll = {}
        for i in range(SafeRepr.maxcollection[0]):
            dcoll[str(i) * SafeRepr.maxstring_outer] = coll
        text = self.saferepr(dcoll)
        # try:
        #    text_repr = repr(dcoll)
        # except MemoryError:
        #    print('Memory error raised while creating repr of test data')
        #    text_repr = ''
        # print('len(SafeRepr()(dcoll)) = ' + str(len(text)) +
        #      ', len(repr(coll)) = ' + str(len(text_repr)))

        assert len(text) < 8192


class TestStrings(SafeReprTestBase):

    def test_str_small(self):
        value = 'A' * 5

        self.assert_unchanged(value, "'AAAAA'")
        self.assert_unchanged([value], "['AAAAA']")

    def test_str_large(self):
        value = 'A' * (SafeRepr.maxstring_outer + 10)

        self.assert_shortened(value,
                              "'" + 'A' * 43690 + "..." + 'A' * 21845 + "'")
        self.assert_shortened([value], "['AAAAAAAAAAAAAAAAAAAA...AAAAAAAAAA']")

    def test_str_largest_unchanged(self):
        value = 'A' * (SafeRepr.maxstring_outer)

        self.assert_unchanged(value, "'" + 'A' * 65536 + "'")

    def test_str_smallest_changed(self):
        value = 'A' * (SafeRepr.maxstring_outer + 1)

        self.assert_shortened(value,
                              "'" + 'A' * 43690 + "..." + 'A' * 21845 + "'")

    def test_str_list_largest_unchanged(self):
        value = 'A' * (SafeRepr.maxstring_inner)

        self.assert_unchanged([value], "['AAAAAAAAAAAAAAAAAAAAAAAAAAAAAA']")

    def test_str_list_smallest_changed(self):
        value = 'A' * (SafeRepr.maxstring_inner + 1)

        self.assert_shortened([value], "['AAAAAAAAAAAAAAAAAAAA...AAAAAAAAAA']")

    @pytest.mark.skipif(sys.version_info > (3, 0), reason='Py2 specific test')
    def test_unicode_small(self):
        value = u'A' * 5

        self.assert_unchanged(value, "u'AAAAA'")
        self.assert_unchanged([value], "[u'AAAAA']")

    @pytest.mark.skipif(sys.version_info > (3, 0), reason='Py2 specific test')
    def test_unicode_large(self):
        value = u'A' * (SafeRepr.maxstring_outer + 10)

        self.assert_shortened(value,
                              "u'" + 'A' * 43690 + "..." + 'A' * 21845 + "'")
        self.assert_shortened([value], "[u'AAAAAAAAAAAAAAAAAAAA...AAAAAAAAAA']")

    @pytest.mark.skipif(sys.version_info < (3, 0), reason='Py3 specific test')
    def test_bytes_small(self):
        value = b'A' * 5

        self.assert_unchanged(value, "b'AAAAA'")
        self.assert_unchanged([value], "[b'AAAAA']")

    @pytest.mark.skipif(sys.version_info < (3, 0), reason='Py3 specific test')
    def test_bytes_large(self):
        value = b'A' * (SafeRepr.maxstring_outer + 10)

        self.assert_shortened(value,
                              "b'" + 'A' * 43690 + "..." + 'A' * 21845 + "'")
        self.assert_shortened([value], "[b'AAAAAAAAAAAAAAAAAAAA...AAAAAAAAAA']")

    # @pytest.mark.skip(reason='not written')  # TODO: finish!
    # def test_bytearray_small(self):
    #    raise NotImplementedError
    #
    # @pytest.mark.skip(reason='not written')  # TODO: finish!
    # def test_bytearray_large(self):
    #    raise NotImplementedError


class RawValueTests(SafeReprTestBase):

    def setUp(self):
        super(RawValueTests, self).setUp()
        self.saferepr.raw_value = True

    def test_unicode_raw(self):
        value = u'A\u2000' * 10000
        self.assert_saferepr(value, value)

    def test_bytes_raw(self):
        value = b'A' * 10000
        self.assert_saferepr(value, value.decode('ascii'))

    def test_bytearray_raw(self):
        value = bytearray(b'A' * 5)
        self.assert_saferepr(value, value.decode('ascii'))

# class TestNumbers(SafeReprTestBase):
#
#     @pytest.mark.skip(reason='not written')  # TODO: finish!
#     def test_int(self):
#         raise NotImplementedError
#
#     @pytest.mark.skip(reason='not written')  # TODO: finish!
#     def test_float(self):
#         raise NotImplementedError
#
#     @pytest.mark.skip(reason='not written')  # TODO: finish!
#     def test_complex(self):
#         raise NotImplementedError


class ContainerBase(object):

    CLASS = None
    LEFT = None
    RIGHT = None

    @property
    def info(self):
        try:
            return self._info
        except AttributeError:
            for info in SafeRepr.collection_types:
                ctype, _, _, _ = info
                if self.CLASS is ctype:
                    type(self)._info = info
                    return info
            else:
                raise TypeError('unsupported')

    def _combine(self, items, prefix, suffix, large):
        contents = ', '.join(str(item) for item in items)
        if large:
            contents += ', ...'
        return prefix + contents + suffix

    def combine(self, items, large=False):
        if self.LEFT is None:
            pytest.skip('unsupported')
        return self._combine(items, self.LEFT, self.RIGHT, large=large)

    def combine_nested(self, depth, items, large=False):
        _, _prefix, _suffix, comma = self.info
        prefix = _prefix * (depth + 1)
        if comma:
            suffix = _suffix + ("," + _suffix) * depth
        else:
            suffix = _suffix * (depth + 1)
        # print("ctype = " + ctype.__name__ + ", maxcollection[" +
        #      str(i) + "] == " + str(SafeRepr.maxcollection[i]))
        return self._combine(items, prefix, suffix, large=large)

    def test_large_flat(self):
        c1 = self.CLASS(range(SafeRepr.maxcollection[0] * 2))
        items = range(SafeRepr.maxcollection[0] - 1)
        c1_expect = self.combine(items, large=True)

        self.assert_shortened(c1, c1_expect)

    def test_large_nested(self):
        c1 = self.CLASS(range(SafeRepr.maxcollection[0] * 2))
        c1_items = range(SafeRepr.maxcollection[1] - 1)
        c1_expect = self.combine(c1_items, large=True)

        c2 = self.CLASS(c1 for _ in range(SafeRepr.maxcollection[0] * 2))
        items = (c1_expect for _ in range(SafeRepr.maxcollection[0] - 1))
        c2_expect = self.combine(items, large=True)

        self.assert_shortened(c2, c2_expect)

    # @pytest.mark.skip(reason='not written')  # TODO: finish!
    # def test_empty(self):
    #     raise NotImplementedError
    #
    # @pytest.mark.skip(reason='not written')  # TODO: finish!
    # def test_subclass(self):
    #     raise NotImplementedError

    def test_boundary(self):
        items1 = range(SafeRepr.maxcollection[0] - 1)
        items2 = range(SafeRepr.maxcollection[0])
        items3 = range(SafeRepr.maxcollection[0] + 1)
        c1 = self.CLASS(items1)
        c2 = self.CLASS(items2)
        c3 = self.CLASS(items3)
        expected1 = self.combine(items1)
        expected2 = self.combine(items2[:-1], large=True)
        expected3 = self.combine(items3[:-2], large=True)

        self.assert_unchanged(c1, expected1)
        self.assert_shortened(c2, expected2)
        self.assert_shortened(c3, expected3)

    def test_nested(self):
        ctype = self.CLASS
        for i in range(1, len(SafeRepr.maxcollection)):
            items1 = range(SafeRepr.maxcollection[i] - 1)
            items2 = range(SafeRepr.maxcollection[i])
            items3 = range(SafeRepr.maxcollection[i] + 1)
            c1 = self.CLASS(items1)
            c2 = self.CLASS(items2)
            c3 = self.CLASS(items3)
            for _j in range(i):
                c1, c2, c3 = ctype((c1,)), ctype((c2,)), ctype((c3,))
            expected1 = self.combine_nested(i, items1)
            expected2 = self.combine_nested(i, items2[:-1], large=True)
            expected3 = self.combine_nested(i, items3[:-2], large=True)

            self.assert_unchanged(c1, expected1)
            self.assert_shortened(c2, expected2)
            self.assert_shortened(c3, expected3)


class TestTuples(ContainerBase, SafeReprTestBase):

    CLASS = tuple
    LEFT = '('
    RIGHT = ')'


class TestLists(ContainerBase, SafeReprTestBase):

    CLASS = list
    LEFT = '['
    RIGHT = ']'

    def test_directly_recursive(self):
        value = [1, 2]
        value.append(value)

        self.assert_unchanged(value, '[1, 2, [...]]')

    def test_indirectly_recursive(self):
        value = [1, 2]
        value.append([value])

        self.assert_unchanged(value, '[1, 2, [[...]]]')


class TestFrozensets(ContainerBase, SafeReprTestBase):

    CLASS = frozenset


class TestSets(ContainerBase, SafeReprTestBase):

    CLASS = set
    if PY_VER != 2:
        LEFT = '{'
        RIGHT = '}'

    def test_nested(self):
        pytest.skip('unsupported')

    def test_large_nested(self):
        pytest.skip('unsupported')


class TestDicts(SafeReprTestBase):

    def test_large_key(self):
        value = {
            'a' * SafeRepr.maxstring_inner * 3: '',
        }

        self.assert_shortened_regex(value, r"{'a+\.\.\.a+': ''}")

    def test_large_value(self):
        value = {
            '': 'a' * SafeRepr.maxstring_inner * 2,
        }

        self.assert_shortened_regex(value, r"{'': 'a+\.\.\.a+'}")

    def test_large_both(self):
        value = {}
        key = 'a' * SafeRepr.maxstring_inner * 2
        value[key] = key

        self.assert_shortened_regex(value, r"{'a+\.\.\.a+': 'a+\.\.\.a+'}")

    def test_nested_value(self):
        d1 = {}
        d1_key = 'a' * SafeRepr.maxstring_inner * 2
        d1[d1_key] = d1_key
        d2 = {d1_key: d1}
        d3 = {d1_key: d2}

        self.assert_shortened_regex(d2, r"{'a+\.\.\.a+': {'a+\.\.\.a+': 'a+\.\.\.a+'}}")  # noqa
        if len(SafeRepr.maxcollection) == 2:
            self.assert_shortened_regex(d3, r"{'a+\.\.\.a+': {'a+\.\.\.a+': {\.\.\.}}}")  # noqa
        else:
            self.assert_shortened_regex(d3, r"{'a+\.\.\.a+': {'a+\.\.\.a+': {'a+\.\.\.a+': 'a+\.\.\.a+'}}}")  # noqa

    def test_empty(self):
        # Ensure empty dicts work
        self.assert_unchanged({}, '{}')

    def test_sorted(self):
        # Ensure dict keys are sorted
        d1 = {}
        d1['c'] = None
        d1['b'] = None
        d1['a'] = None
        if IS_PY36_OR_GREATER:
            self.assert_saferepr(d1, "{'c': None, 'b': None, 'a': None}")
        else:
            self.assert_saferepr(d1, "{'a': None, 'b': None, 'c': None}")

    @pytest.mark.skipif(sys.version_info < (3, 0), reason='Py3 specific test')
    def test_unsortable_keys(self):
        # Ensure dicts with unsortable keys do not crash
        d1 = {}
        for _ in range(100):
            d1[object()] = None
        with pytest.raises(TypeError):
            list(sorted(d1))
        self.saferepr(d1)

    def test_directly_recursive(self):
        value = {1: None}
        value[2] = value

        self.assert_unchanged(value, '{1: None, 2: {...}}')

    def test_indirectly_recursive(self):
        value = {1: None}
        value[2] = {3: value}

        self.assert_unchanged(value, '{1: None, 2: {3: {...}}}')


class TestOtherPythonTypes(SafeReprTestBase):
    # not critical to test:
    #  singletons
    #  <function>
    #  <class>
    #  <iterator>
    #  memoryview
    #  classmethod
    #  staticmethod
    #  property
    #  enumerate
    #  reversed
    #  object
    #  type
    #  super

    # @pytest.mark.skip(reason='not written')  # TODO: finish!
    # def test_file(self):
    #     raise NotImplementedError

    def test_range_small(self):
        range_name = range.__name__
        value = range(1, 42)

        self.assert_unchanged(value, '%s(1, 42)' % (range_name,))

    @pytest.mark.skipif(sys.version_info < (3, 0), reason='Py3 specific test')
    def test_range_large_stop_only(self):
        range_name = range.__name__
        stop = SafeRepr.maxcollection[0]
        value = range(stop)

        self.assert_unchanged(value,
                              '%s(0, %s)' % (range_name, stop))

    def test_range_large_with_start(self):
        range_name = range.__name__
        stop = SafeRepr.maxcollection[0] + 1
        value = range(1, stop)

        self.assert_unchanged(value,
                              '%s(1, %s)' % (range_name, stop))

    # @pytest.mark.skip(reason='not written')  # TODO: finish!
    # def test_named_struct(self):
    #     # e.g. sys.version_info
    #     raise NotImplementedError
    #
    # @pytest.mark.skip(reason='not written')  # TODO: finish!
    # def test_namedtuple(self):
    #     raise NotImplementedError
    #
    # @pytest.mark.skip(reason='not written')  # TODO: finish!
    # @pytest.mark.skipif(sys.version_info < (3, 0), reason='Py3 specific test')
    # def test_SimpleNamespace(self):
    #     raise NotImplementedError


class TestUserDefinedObjects(SafeReprTestBase):

    def test_broken_repr(self):

        class TestClass(object):

            def __repr__(self):
                raise NameError

        value = TestClass()

        with pytest.raises(NameError):
            repr(TestClass())
        self.assert_saferepr(value, object.__repr__(value))

    def test_large(self):

        class TestClass(object):

            def __repr__(self):
                return '<' + 'A' * SafeRepr.maxother_outer * 2 + '>'

        value = TestClass()

        self.assert_shortened_regex(value, r'\<A+\.\.\.A+\>')

    def test_inherit_repr(self):

        class TestClass(dict):
            pass

        value_dict = TestClass()

        class TestClass2(list):
            pass

        value_list = TestClass2()

        self.assert_unchanged(value_dict, '{}')
        self.assert_unchanged(value_list, '[]')

    def test_custom_repr(self):

        class TestClass(dict):

            def __repr__(self):
                return 'MyRepr'

        value1 = TestClass()

        class TestClass2(list):

            def __repr__(self):
                return 'MyRepr'

        value2 = TestClass2()

        self.assert_unchanged(value1, 'MyRepr')
        self.assert_unchanged(value2, 'MyRepr')

    def test_custom_repr_many_items(self):

        class TestClass(list):

            def __init__(self, it=()):
                list.__init__(self, it)

            def __repr__(self):
                return 'MyRepr'

        value1 = TestClass(range(0, 15))
        value2 = TestClass(range(0, 16))
        value3 = TestClass([TestClass(range(0, 10))])
        value4 = TestClass([TestClass(range(0, 11))])

        self.assert_unchanged(value1, 'MyRepr')
        self.assert_shortened(value2, '<TestClass, len() = 16>')
        self.assert_unchanged(value3, 'MyRepr')
        self.assert_shortened(value4, '<TestClass, len() = 1>')

    def test_custom_repr_large_item(self):

        class TestClass(list):

            def __init__(self, it=()):
                list.__init__(self, it)

            def __repr__(self):
                return 'MyRepr'

        value1 = TestClass(['a' * (SafeRepr.maxcollection[1] + 1)])
        value2 = TestClass(['a' * (SafeRepr.maxstring_inner + 1)])

        self.assert_unchanged(value1, 'MyRepr')
        self.assert_shortened(value2, '<TestClass, len() = 1>')


@pytest.mark.skipif(np is None, reason='could not import numpy')
class TestNumpy(SafeReprTestBase):
    # numpy types should all use their native reprs, even arrays
    # exceeding limits.

    def test_int32(self):
        value = np.int32(123)

        self.assert_unchanged(value, repr(value))

    def test_float32(self):
        value = np.float32(123.456)

        self.assert_unchanged(value, repr(value))

    def test_zeros(self):
        value = np.zeros(SafeRepr.maxcollection[0] + 1)

        self.assert_unchanged(value, repr(value))


@pytest.mark.parametrize('params', [
    {'maxother_outer': 20, 'input': "😄😄😄😄😄😄😄😄😄😄😄😄😄😄😄😄😄😄FFFFFFFF", 'output': '😄😄😄😄😄😄😄😄😄😄😄😄😄...FFFFFF'},
    {'maxother_outer': 10, 'input': "😄😄😄😄😄😄😄😄😄😄😄😄😄😄😄😄😄😄FFFFFFFF", 'output': '😄😄😄😄😄😄...FFF'},
    {'maxother_outer': 10, 'input': u"������������FFFFFFFF", 'output': u"������...FFF"},

    # Because we can't return bytes, byte-related tests aren't needed (and str works as it should).
])
@pytest.mark.parametrize('use_str', [True, False])
def test_py3_str_slicing(params, use_str):
    # Note: much simpler in python because __repr__ is required to return str
    safe_repr = SafeRepr()
    safe_repr.locale_preferred_encoding = 'ascii'
    safe_repr.sys_stdout_encoding = params.get('sys_stdout_encoding', 'ascii')

    safe_repr.maxother_outer = params['maxother_outer']

    if not use_str:

        class MyObj(object):

            def __repr__(self):
                return params['input']

        safe_repr_input = MyObj()
    else:
        safe_repr_input = params['input']
    expected_output = params['output']
    computed = safe_repr(safe_repr_input)
    expected = repr(expected_output)
    if use_str:
        expected = repr(expected)
    assert repr(computed) == expected

    # Check that we can json-encode the return.
    assert json.dumps(computed)


def test_raw_bytes():
    safe_repr = SafeRepr()
    safe_repr.raw_value = True
    obj = b'\xed\xbd\xbf\xff\xfe\xfa\xfd'
    raw_value_repr = safe_repr(obj)
    assert isinstance(raw_value_repr, str)  # bytes on py2, str on py3
    assert raw_value_repr == obj.decode('latin1')


def test_raw_unicode():
    safe_repr = SafeRepr()
    safe_repr.raw_value = True
    obj = u'\xed\xbd\xbf\xff\xfe\xfa\xfd'
    raw_value_repr = safe_repr(obj)
    assert isinstance(raw_value_repr, str)  # bytes on py2, str on py3
    assert raw_value_repr == obj


def test_no_repr():

    class MyBytes(object):

        def __init__(self, contents):
            self.contents = contents
            self.errored = None

        def __iter__(self):
            return iter(self.contents)

        def decode(self, encoding):
            self.errored = 'decode called'
            raise RuntimeError('Should not be called.')

        def __repr__(self):
            self.errored = '__repr__ called'
            raise RuntimeError('Should not be called.')

        def __getitem__(self, *args):
            return self.contents.__getitem__(*args)

        def __len__(self):
            return len(self.contents)

    safe_repr = SafeRepr()
    safe_repr.string_types = (MyBytes,)
    safe_repr.bytes = MyBytes
    obj = b'f' * (safe_repr.maxstring_outer * 10)
    my_bytes = MyBytes(obj)
    raw_value_repr = safe_repr(my_bytes)
    assert not my_bytes.errored




if __name__ == '__main__':
    import os
    import sys
    port = int(sys.argv[1])
    root_dirname = os.path.dirname(os.path.dirname(__file__))
    
    if root_dirname not in sys.path:
        sys.path.append(root_dirname)
        
    import pydevd
    print('before pydevd.settrace')
    pydevd.settrace(port=port)
    print('after pydevd.settrace')
    for i in range(2):
        try:
            raise ValueError('not finished')
        except:
            pass
    
    raise ValueError('TEST SUCEEDED!')
    


import os
import sys
import subprocess


def breaknow():
    print('break here')


if '--fork-in-subprocess' in sys.argv:
    if sys.platform == 'win32':
        popen = subprocess.Popen([sys.executable, __file__, '--forked'])
        pid = popen.pid
    else:
        pid = os.fork()
    try:
        ppid = os.getppid()
    except:
        ppid = '<unknown>'
    print('currently in pid: %s, ppid: %s' % (os.getpid(), ppid))
    print('os.fork returned', pid)
    breaknow()

elif '--forked' in sys.argv:
    try:
        ppid = os.getppid()
    except:
        ppid = '<unknown>'
    print('currently in pid: %s, ppid: %s' % (os.getpid(), ppid))
    breaknow()

elif '--fork-in-subprocess' not in sys.argv:
    out = subprocess.check_output([sys.executable, __file__, '--fork-in-subprocess'])
    breaknow()
    print('\n\nin pid %s, output from subprocess.run:\n%s' % (os.getpid(), out.decode('utf-8')))
    print('TEST SUCEEDED!')



import asyncio


async def main():
    import sys
    import os
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))

    from not_my_code import not_my_coroutine
    try:
        await not_my_coroutine.call1(call2)  # stop here 2
    finally:
        print('finish main')


async def call2():
    try:
        raise RuntimeError()  # stop here 1
    except RuntimeError:
        pass


if __name__ == '__main__':
    try:
        if hasattr(asyncio, 'run'):
            print('using asyncio.run')
            asyncio.run(main())  # stop here 3a
        else:
            print('using loop.run_until_complete')
            loop = asyncio.get_event_loop()
            loop.run_until_complete(main())  # stop here 3b
            loop.close()
    finally:
        print('TEST SUCEEDED!')
